from __future__ import unicode_literals

from prompt_toolkit.utils import get_cwidth
from collections import defaultdict, namedtuple
from pygments.token import Token
import six

__all__ = (
    'Point',
    'Size',
    'Screen',
    'Char',
)


Point = namedtuple('Point', 'y x')
Size = namedtuple('Size', 'rows columns')


class Char(object):
    __slots__ = ('char', 'token', 'width')

    # If we end up having one of these special control sequences in the input string,
    # we should display them as follows:
    # Usually this happens after a "quoted insert".
    display_mappings = {
        '\x00': '^@',  # Control space
        '\x01': '^A',
        '\x02': '^B',
        '\x03': '^C',
        '\x04': '^D',
        '\x05': '^E',
        '\x06': '^F',
        '\x07': '^G',
        '\x08': '^H',
        '\x09': '^I',
        '\x0a': '^J',
        '\x0b': '^K',
        '\x0c': '^L',
        '\x0d': '^M',
        '\x0e': '^N',
        '\x0f': '^O',
        '\x10': '^P',
        '\x11': '^Q',
        '\x12': '^R',
        '\x13': '^S',
        '\x14': '^T',
        '\x15': '^U',
        '\x16': '^V',
        '\x17': '^W',
        '\x18': '^X',
        '\x19': '^Y',
        '\x1a': '^Z',
        '\x1b': '^[',  # Escape
        '\x1c': '^\\',
        '\x1d': '^]',
        '\x1f': '^_',
        '\x7f': '^?',  # Control backspace
    }

    def __init__(self, char=' ', token=Token):
        # If this character has to be displayed otherwise, take that one.
        char = self.display_mappings.get(char, char)

        self.char = char
        self.token = token

        # Calculate width. (We always need this, so better to store it directly
        # as a member for performance.)
        self.width = get_cwidth(char)

    def __repr__(self):
        return 'Char(%r, %r)' % (self.char, self.token)


class Screen(object):
    """
    Two dimentional buffer for the output.
    """
    def __init__(self, width, default_char=None):
        if default_char:
            self._buffer = defaultdict(lambda: defaultdict(lambda: default_char))
        else:
            self._buffer = defaultdict(lambda: defaultdict(Char))

        self.width = width

        #: Position of the cursor.
        self.cursor_position = Point(y=0, x=0)

        #: (Optional) Where to position the menu. E.g. at the start of a completion.
        #: (We can't use the cursor position, because we don't want the
        #: completion menu to change its position when we browse through all the
        #: completions.)
        self.menu_position = None

        #: Currently used height of the screen. This will increase when data is
        #: written to the screen.
        self.current_height = 1

    def write_at_position(self, data, width, margin=None):  # XXX: rename to write_data()
        """
        Write data at :class:`WritePosition`.
        """
        buffer = self._buffer
        x = 0
        y = 0
        max_x = x + width
        index = 0
        line_number = 0
        requires_line_feed = True
        indexes_to_pos = {}  # Map input positions to (x, y) coordinates.

        def do_line_feed(x, number):
            """
            Insert left margin. (line numbering or whatever provider by the
            'margin' function.)
            """
            if margin is not None:
                for token, text in margin(number):
                    for char in text:
                        char_obj = Char(char, token)
                        char_width = char_obj.width
                        buffer[y][x] = char_obj

                        # Store '0' in the second cell of double width characters.
                        if char_width > 1:
                            buffer[y][x+1] = Char(six.unichr(0))

                        # Move position
                        x += char_width
            return x

        for token, text in data:
            for char in text:
                # Line feed.
                if requires_line_feed:
                    x = do_line_feed(x, line_number)
                    requires_line_feed = False

                char_obj = Char(char, token)
                char_width = char_obj.width

                # In case there is no more place left at this line, go first to the
                # following line. (Also in case of double-width characters.)
                if x + char_width > max_x and char != '\n':
                    y += 1
                    x = 0
                    x = do_line_feed(x, None)

                # Keep mapping of index to position.
                indexes_to_pos[index] = (x, y)

                # Insertion of newline
                if char == '\n':
                    y += 1
                    x = 0
                    requires_line_feed = True
                    line_number += 1

                # Insertion of a 'visible' character.
                else:
                    buffer_y = buffer[y]
                    buffer_y[x] = char_obj

                    # When we have a double width character, store this byte in the
                    # second cell. So that if this character gets deleted afterwarsd,
                    # the ``output_screen_diff`` will notice that this byte is also
                    # gone and redraw both cells.
                    if char_width > 1:
                        buffer_y[x+1] = Char(six.unichr(0))

                    # Move position
                    x += char_width

                index += 1

        self.current_height = max(self.current_height, y+1)

        return indexes_to_pos


class WritePosition(object):
    def __init__(self, xpos, ypos, width, height, extended_height=None):
        assert height >= 0
        assert extended_height is None or extended_height >= 0
        assert width >= 0
        assert xpos >= 0
        assert ypos >= 0

        self.xpos = xpos
        self.ypos = ypos
        self.width = width
        self.height = height
        self.extended_height = extended_height or height

    def __repr__(self):
        return 'WritePosition(%r, %r, %r %r, %r)' % (
            self.xpos, self.ypos, self.width, self.height, self.extended_height)



"""
Parser for VT100 input stream.
"""
from __future__ import unicode_literals

import os
import re
import six
import termios
import tty

from ..keys import Keys
from ..key_binding.input_processor import KeyPress

__all__ = (
    'InputStream',
    'raw_mode',
    'cooked_mode',
)

_DEBUG_RENDERER_INPUT = False
_DEBUG_RENDERER_INPUT_FILENAME = 'prompt-toolkit-render-input.log'


# Regex matching any CPR response
# (Note that we use '\Z' instead of '$', because '$' could include a trailing
# newline.)
_cpr_response_re = re.compile('^' + re.escape('\x1b[') + r'\d+;\d+R\Z')

# Regex matching any valid prefix of a CPR response.
# (Note that it doesn't contain the last character, the 'R'. The prefix has to
# be shorter.)
_cpr_response_prefix_re = re.compile('^' + re.escape('\x1b[') + r'[\d;]*\Z')


class _Flush(object):
    """ Helper object to indicate flush operation to the parser. """
    pass


class InputStream(object):
    """
    Parser for VT100 input stream.

    Feed the data through the `feed` method and the correct callbacks of the
    `input_processor` will be called.

    ::

        h = InputStreamHandler()
        i = InputStream(h)
        i.feed('data\x01...')

    :attr input_processor: :class:`~prompt_toolkit.key_binding.InputProcessor` instance.
    """
    # Lookup table of ANSI escape sequences for a VT100 terminal
    # Hint: in order to know what sequences your terminal writes to stdin, run
    #       "od -c" and start typing.
    mappings = {
        '\x1b': Keys.Escape,

        '\x00': Keys.ControlSpace,  # Control-Space (Also for Ctrl-@)
        '\x01': Keys.ControlA,  # Control-A (home)
        '\x02': Keys.ControlB,  # Control-B (emacs cursor left)
        '\x03': Keys.ControlC,  # Control-C (interrupt)
        '\x04': Keys.ControlD,  # Control-D (exit)
        '\x05': Keys.ControlE,  # Contrel-E (end)
        '\x06': Keys.ControlF,  # Control-F (cursor forward)
        '\x07': Keys.ControlG,  # Control-G
        '\x08': Keys.ControlH,  # Control-H (8) (Identical to '\b')
        '\x09': Keys.ControlI,  # Control-I (9) (Identical to '\t')
        '\x0a': Keys.ControlJ,  # Control-J (10) (Identical to '\n')
        '\x0b': Keys.ControlK,  # Control-K (delete until end of line; vertical tab)
        '\x0c': Keys.ControlL,  # Control-L (clear; form feed)
        '\x0d': Keys.ControlM,  # Control-M (13) (Identical to '\r')
        '\x0e': Keys.ControlN,  # Control-N (14) (history forward)
        '\x0f': Keys.ControlO,  # Control-O (15)
        '\x10': Keys.ControlP,  # Control-P (16) (history back)
        '\x11': Keys.ControlQ,  # Control-Q
        '\x12': Keys.ControlR,  # Control-R (18) (reverse search)
        '\x13': Keys.ControlS,  # Control-S (19) (forward search)
        '\x14': Keys.ControlT,  # Control-T
        '\x15': Keys.ControlU,  # Control-U
        '\x16': Keys.ControlV,  # Control-V
        '\x17': Keys.ControlW,  # Control-W
        '\x18': Keys.ControlX,  # Control-X
        '\x19': Keys.ControlY,  # Control-Y (25)
        '\x1a': Keys.ControlZ,  # Control-Z

        '\x1c': Keys.ControlBackslash,  # Both Control-\ and Ctrl-|
        '\x1d': Keys.ControlSquareClose,  # Control-]
        '\x1e': Keys.ControlCircumflex,  # Control-^
        '\x1f': Keys.ControlUnderscore,  # Control-underscore (Also for Ctrl-hypen.)
        '\x7f': Keys.Backspace,  # (127) Backspace
        '\x1b[A': Keys.Up,
        '\x1b[B': Keys.Down,
        '\x1b[C': Keys.Right,
        '\x1b[D': Keys.Left,
        '\x1b[H': Keys.Home,
        '\x1bOH': Keys.Home,
        '\x1b[F': Keys.End,
        '\x1bOF': Keys.End,
        '\x1b[3~': Keys.Delete,
        '\x1b[3;2~': Keys.ShiftDelete,  # xterm, gnome-terminal.
        '\x1b[1~': Keys.Home,  # tmux
        '\x1b[4~': Keys.End,  # tmux
        '\x1b[5~': Keys.PageUp,
        '\x1b[6~': Keys.PageDown,
        '\x1b[7~': Keys.Home,  # xrvt
        '\x1b[8~': Keys.End,  # xrvt
        '\x1b[Z': Keys.BackTab,  # shift + tab

        '\x1bOP': Keys.F1,
        '\x1bOQ': Keys.F2,
        '\x1bOR': Keys.F3,
        '\x1bOS': Keys.F4,
        '\x1b[11~': Keys.F1,  # rxvt-unicode
        '\x1b[12~': Keys.F2,  # rxvt-unicode
        '\x1b[13~': Keys.F3,  # rxvt-unicode
        '\x1b[14~': Keys.F4,  # rxvt-unicode
        '\x1b[15~': Keys.F5,
        '\x1b[17~': Keys.F6,
        '\x1b[18~': Keys.F7,
        '\x1b[19~': Keys.F8,
        '\x1b[20~': Keys.F9,
        '\x1b[21~': Keys.F10,
        '\x1b[23~': Keys.F11,
        '\x1b[24~': Keys.F12,
        '\x1b[25~': Keys.F13,
        '\x1b[26~': Keys.F14,
        '\x1b[28~': Keys.F15,
        '\x1b[29~': Keys.F16,
        '\x1b[31~': Keys.F17,
        '\x1b[32~': Keys.F18,
        '\x1b[33~': Keys.F19,
        '\x1b[34~': Keys.F20,
        '\x1b[1;5A': Keys.ControlUp,     # Cursor Mode
        '\x1b[1;5B': Keys.ControlDown,   # Cursor Mode
        '\x1b[1;5C': Keys.ControlRight,  # Cursor Mode
        '\x1b[1;5D': Keys.ControlLeft,   # Cursor Mode
        '\x1bOA': Keys.ControlUp,       # Application Mode (tmux)
        '\x1bOB': Keys.ControlDown,     # Application Mode (tmux)
        '\x1bOC': Keys.ControlRight,    # Application Mode (tmux)
        '\x1bOD': Keys.ControlLeft,     # Application Mode (tmux)

        '\x1b[5A': Keys.ControlUp,
        '\x1b[5B': Keys.ControlDown,
        '\x1b[5C': Keys.ControlRight,
        '\x1b[5D': Keys.ControlLeft,

        # Meta + arrow keys. Several terminals handle this differently.
        # The following sequences are for xterm and gnome-terminal.
        #     (Iterm sends ESC followed by the normal arrow_up/down/left/right
        #     sequences, and the OSX Terminal sends ESCb and ESCf for "alt
        #     arrow_left" and "alt arrow_right." We don't handle these
        #     explicitely, in here, because would could not distinguesh between
        #     pressing ESC (to go to Vi navigation mode), followed by just the
        #     'b' or 'f' key. These combinations are handled in
        #     the input processor.)
        '\x1b[1;3D': (Keys.Escape, Keys.Left),
        '\x1b[1;3C': (Keys.Escape, Keys.Right),
        '\x1b[1;3A': (Keys.Escape, Keys.Up),
        '\x1b[1;3B': (Keys.Escape, Keys.Down),
    }

    def __init__(self, input_processor):
        self._input_processor = input_processor
        self.reset()

        if _DEBUG_RENDERER_INPUT:
            self.LOG = open(_DEBUG_RENDERER_INPUT_FILENAME, 'ab')

    def reset(self, request=False):
        self._start_parser()

    def _start_parser(self):
        """
        Start the parser coroutine.
        """
        self._input_parser = self._input_parser_generator()
        self._input_parser.send(None)

    def _get_matches(self, prefix):
        """
        Return the keys that map to this prefix.
        """
        # (hard coded) If we match a CPR response, return Keys.CPRResponse.
        # (This one doesn't fit in the mappings, because it contains
        # integer variables.)
        if _cpr_response_re.match(prefix):
            return [Keys.CPRResponse]

        # Otherwise, use the mappings.
        return [v for k, v in self.mappings.items() if k == prefix]

    def _is_prefix_of_longer_match(self, prefix):
        """
        True when there is any key that start with this characters.
        """
        # (hard coded) If this could be a prefix of a CPR response, return
        # True.
        if _cpr_response_prefix_re.match(prefix):
            return True

        # If this could be a prefix of anything else, also return True.
        return any(v for k, v in self.mappings.items() if k.startswith(prefix) and k != prefix)

    def _input_parser_generator(self):
        """
        Coroutine (state machine) for the input parser.
        """
        prefix = ''
        retry = False
        flush = False

        while True:
            flush = False

            if retry:
                retry = False
            else:
                # Get next character.
                c = yield

                if c == _Flush:
                    flush = True
                else:
                    prefix += c

            # If we have some data, check for matches.
            if prefix:
                is_prefix_of_longer_match = self._is_prefix_of_longer_match(prefix)
                matches = self._get_matches(prefix)

                # Exact matches found, call handlers..
                if (flush or not is_prefix_of_longer_match) and matches:
                    self._call_handler(matches[0], prefix)
                    prefix = ''

                # No exact matches found.
                elif (flush or not is_prefix_of_longer_match) and not matches:
                    found = False
                    retry = True

                    # Loop over the input, try the longest match first and
                    # shift.
                    for i in range(len(prefix), 0, -1):
                        matches = self._get_matches(prefix[:i])
                        if matches:
                            self._call_handler(matches[0], prefix[:i])
                            prefix = prefix[i:]
                            found = True

                    if not found:
                        self._call_handler(prefix[0], prefix[0])
                        prefix = prefix[1:]

    def _call_handler(self, key, insert_text):
        """
        Callback to handler.
        """
        if isinstance(key, tuple):
            for k in key:
                self._call_handler(k, insert_text)
        else:
            self._input_processor.feed_key(KeyPress(key, insert_text))

    def feed(self, data):
        """
        Feed the input stream.
        """
        assert isinstance(data, six.text_type)

        if _DEBUG_RENDERER_INPUT:
            self.LOG.write(repr(data).encode('utf-8') + b'\n')
            self.LOG.flush()

        for c in data:
            self._input_parser.send(c)

    def flush(self):
        """
        Flush the buffer of the input stream.

        This will allow us to handle the escape key (or maybe meta) sooner.
        The input received by the escape key is actually the same as the first
        characters of e.g. Arrow-Up, so without knowing what follows the escape
        sequence, we don't know whether escape has been pressed, or whether
        it's something else. This flush function should be called after a
        timeout, and processes everything that's still in the buffer as-is, so
        without assuming any characters will folow.
        """
        self._input_parser.send(_Flush)

    def feed_and_flush(self, data):
        """
        Wrapper around ``feed`` and ``flush``.
        """
        self.feed(data)
        self.flush()


class raw_mode(object):
    """
    ::

        with raw_mode(stdin):
            ''' the pseudo-terminal stdin is now used in raw mode '''
    """
    def __init__(self, fileno):
        self.fileno = fileno
        self.attrs_before = termios.tcgetattr(fileno)

    def __enter__(self):
        # NOTE: On os X systems, using pty.setraw() fails. Therefor we are using this:
        newattr = termios.tcgetattr(self.fileno)
        newattr[tty.LFLAG] = self._patch(newattr[tty.LFLAG])
        termios.tcsetattr(self.fileno, termios.TCSANOW, newattr)

        # Put the terminal in cursor mode. (Instead of application mode.)
        os.write(self.fileno, b'\x1b[?1l')

    def _patch(self, attrs):
        return attrs & ~(termios.ECHO | termios.ICANON | termios.IEXTEN | termios.ISIG)

    def __exit__(self, *a, **kw):
        termios.tcsetattr(self.fileno, termios.TCSANOW, self.attrs_before)

        # # Put the terminal in application mode.
        # self._stdout.write('\x1b[?1h')


class cooked_mode(raw_mode):
    """
    (The opposide of ``raw_mode``::

        with cooked_mode(stdin):
            ''' the pseudo-terminal stdin is now used in cooked mode. '''
    """
    def _patch(self, attrs):
        return attrs | (termios.ECHO | termios.ICANON | termios.IEXTEN | termios.ISIG)



from __future__ import unicode_literals

from prompt_toolkit.utils import get_cwidth

__all__ = (
    'token_list_len',
    'token_list_width',
    'explode_tokens',
)


def token_list_len(tokenlist):
    """
    Return the amount of characters in this token list.

    :param tokenlist: List of (token, text) tuples.
    """
    return sum(len(v) for k, v in tokenlist)


def token_list_width(tokenlist):
    """
    Return the character width of this token list.
    (Take double width characters into account.)

    :param tokenlist: List of (token, text) tuples.
    """
    return sum(get_cwidth(c) for k, word in tokenlist for c in word)


def explode_tokens(tokenlist):
    """
    Turn a list of (token, text) tuples into another list where each string is
    exactly one character.

    :param tokenlist: List of (token, text) tuples.
    """
    result = []

    for token, string in tokenlist:
        for c in string:
            result.append((token, c))

    return result



"""
:class:`KeyBindingManager` is a utility (or shortcut) for loading all the key
bindings in a key binding registry, with a logic set of filters to quickly to
quickly change from Vi to Emacs key bindings at runtime.

You don't have to use this, but it's practical.

Usage::

    manager = KeyBindingManager()
    cli = CommandLineInterface(key_bindings_registry=manager.registry)
    manager.enable_vi_mode = True
"""
from __future__ import unicode_literals
from prompt_toolkit.filters import Filter
from prompt_toolkit.key_binding.registry import Registry
from prompt_toolkit.key_binding.vi_state import ViState
from prompt_toolkit.key_binding.bindings.emacs import load_emacs_bindings, load_emacs_system_bindings, load_emacs_search_bindings, load_emacs_open_in_editor_bindings
from prompt_toolkit.key_binding.bindings.vi import load_vi_bindings, load_vi_system_bindings, load_vi_search_bindings, load_vi_open_in_editor_bindings

__all__ = (
    'KeyBindingManager',
)


class ManagerFilter(Filter):
    def __init__(self, manager):
        self.manager = manager


class ViModeEnabled(ManagerFilter):
    def __call__(self, cli):
        return self.manager.enable_vi_mode


class SystemPromptEnabled(ManagerFilter):
    def __call__(self, cli):
        return self.manager.enable_system_prompt


class SearchEnabled(ManagerFilter):
    def __call__(self, cli):
        return self.manager.enable_search


class OpenInEditorEnabled(ManagerFilter):
    def __call__(self, cli):
        return self.manager.enable_open_in_editor


class KeyBindingManager(object):
    def __init__(self, registry=None, enable_vi_mode=False,
                 enable_system_prompt=False, enable_search=True,
                 enable_open_in_editor=False):

        self.registry = registry or Registry()

        # Flags. You can change these anytime.
        self.enable_vi_mode = enable_vi_mode
        self.enable_system_prompt = enable_system_prompt
        self.enable_search = enable_search
        self.enable_open_in_editor = enable_open_in_editor

        # Create set of filters to enable/disable sets of key bindings at
        # runtime.
        vi_mode_enabled = ViModeEnabled(self)
        emacs_mode_enabled = ~ vi_mode_enabled
        system_prompt_enabled = SystemPromptEnabled(self)
        search_enabled = SearchEnabled(self)
        open_in_editor_enabled = OpenInEditorEnabled(self)

        # Vi state. (Object to keep track of in which Vi mode we are.)
        self.vi_state = ViState()

        # Load all bindings in the registry with the correct filters.
        load_emacs_bindings(self.registry, emacs_mode_enabled)
        load_emacs_open_in_editor_bindings(
            self.registry, emacs_mode_enabled & open_in_editor_enabled)
        load_emacs_search_bindings(
            self.registry, emacs_mode_enabled & search_enabled)
        load_emacs_system_bindings(
            self.registry, emacs_mode_enabled & system_prompt_enabled)

        load_vi_bindings(self.registry, self.vi_state, vi_mode_enabled)
        load_vi_open_in_editor_bindings(
            self.registry, self.vi_state,
            vi_mode_enabled & open_in_editor_enabled)
        load_vi_search_bindings(
            self.registry, self.vi_state,
            vi_mode_enabled & search_enabled)
        load_vi_system_bindings(
            self.registry, self.vi_state,
            vi_mode_enabled & system_prompt_enabled)

    def reset(self):
        self.vi_state.reset()



from __future__ import unicode_literals

from ..utils import EventHook
import threading

__all__ = (
    'BaseEventLoop',
)


class BaseEventLoop(object):
    #: When to trigger the `onInputTimeout` event.
    input_timeout = .5

    def __init__(self, input_processor, stdin):
        self.stdin = stdin
        self.input_processor = input_processor

        #: Called when there is no input for x seconds.
        #:   At this event, you can for instance start a background thread to
        #:   generate information about the input. E.g. get the code signature
        #:   of the function below the cursor position in the case of a REPL.
        self.onInputTimeout = EventHook()

        self.closed = False

    def loop(self):
        raise NotImplementedError

    def loop_coroutine(self):
        raise NotImplementedError

    def close(self):
        self.closed = True

    def run_in_executor(self, callback):
        """
        Run a long running function in a background thread.
        (This is recommended for code that could block the `read_input` event
        loop.)
        Similar to Twisted's ``deferToThread``.
        """
        threading.Thread(target=callback).start()

    def call_from_executor(self, callback):
        raise NotImplementedError



"""
Container for the layout.
(Containers can contain other containers or user interface controls.)
"""
from __future__ import unicode_literals

from six import with_metaclass
from abc import ABCMeta, abstractmethod

from .screen import Point, WritePosition
from .dimension import LayoutDimension, sum_layout_dimensions, max_layout_dimensions
from .controls import UIControl
from prompt_toolkit.filters import Filter

__all__ = (
    'HSplit',
    'VSplit',
    'FloatContainer',
    'Float',
    'Window',
)


class Layout(with_metaclass(ABCMeta, object)):
    """
    Base class for user interface layout.
    """
    @abstractmethod
    def reset(self):
        pass

    @abstractmethod
    def width(self, cli):  # XXX: rename to preferred_width
        # Should return a LayoutDimension
        pass

    @abstractmethod  # XXX: rename to preferred_height
    def height(self, cli, width):
        # Should return a LayoutDimension
        pass

    @abstractmethod
    def write_to_screen(self, cli, screen, write_position):
        pass


class HSplit(Layout):
    """
    Several layouts, one stacked above/under the other.
    """
    def __init__(self, children):
        assert all(isinstance(c, Layout) for c in children)
        self.children = children

    def width(self, cli):
        dimensions = [c.width(cli) for c in self.children]
        return max_layout_dimensions(dimensions)

    def height(self, cli, width):
        dimensions = [c.height(cli, width) for c in self.children]
        return sum_layout_dimensions(dimensions)

    def reset(self):
        for c in self.children:
            c.reset()

    def write_to_screen(self, cli, screen, write_position):
        """
        Render the prompt to a `Screen` instance.

        :param screen: The :class:`Screen` class into which we write the output.
        """
        # Calculate heights.
        dimensions = [c.height(cli, write_position.width) for c in self.children]
        sum_dimensions = sum_layout_dimensions(dimensions)

        # If there is not enough space for both.
        # Don't do anything. (TODO: show window to small message.)
        if sum_dimensions.min > write_position.extended_height:
            return

        # Find optimal sizes. (Start with minimal size, increase until we cover
        # the whole height.)
        sizes = [d.min for d in dimensions]

        i = 0
        while sum(sizes) < min(write_position.extended_height, sum_dimensions.preferred):
            # Increase until we meet at least the 'preferred' size.
            if sizes[i] < dimensions[i].preferred:
                sizes[i] += 1
            i = (i + 1) % len(sizes)

        if not any([cli.is_returning, cli.is_exiting, cli.is_aborting]):
            while sum(sizes) < min(write_position.height, sum_dimensions.max):
                # Increase until we use all the available space. (or until "max")
                if sizes[i] < dimensions[i].max:
                    sizes[i] += 1
                i = (i + 1) % len(sizes)

        # Draw child panes.
        ypos = write_position.ypos
        xpos = write_position.xpos
        width = write_position.width

        for s, c in zip(sizes, self.children):
            c.write_to_screen(cli, screen, WritePosition(xpos, ypos, width, s))
            ypos += s


class VSplit(Layout):
    """
    Several layouts, one stacked left/right of the other.
    """
    def __init__(self, children):
        assert all(isinstance(c, Layout) for c in children)
        self.children = children

    def width(self, cli):
        dimensions = [c.width(cli) for c in self.children]
        return sum_layout_dimensions(dimensions)

    def height(self, cli, width):
        sizes = self._divide_widths(cli, width)
        if sizes is None:
            return LayoutDimension()
        else:
            dimensions = [c.height(cli, s) for s, c in zip(sizes, self.children)]
            return max_layout_dimensions(dimensions)

    def reset(self):
        for c in self.children:
            c.reset()

    def _divide_widths(self, cli, width):
        """
        Return the widths for all columns.
        Or None when there is not enough space.
        """
        # Calculate widths.
        dimensions = [c.width(cli) for c in self.children]
        sum_dimensions = sum_layout_dimensions(dimensions)

        # If there is not enough space for both.
        # Don't do anything. (TODO: show window too small message.)
        if sum_dimensions.min > width:
            return

        # TODO: like HSplit, first increase until the "preferred" size.

        # Find optimal sizes. (Start with minimal size, increase until we cover
        # the whole height.)
        sizes = [d.min for d in dimensions]
        i = 0
        while sum(sizes) < min(width, sum_dimensions.max):
            if sizes[i] < dimensions[i].max:
                sizes[i] += 1
            i = (i + 1) % len(sizes)

        return sizes

    def write_to_screen(self, cli, screen, write_position):
        """
        Render the prompt to a `Screen` instance.

        :param screen: The :class:`Screen` class into which we write the output.
        """
        sizes = self._divide_widths(cli, write_position.width)

        if sizes is None:
            return

        # Calculate heights, take the largest possible, but not larger than write_position.extended_height.
        heights = [child.height(cli, width).preferred for width, child in zip(sizes, self.children)]
        height = max(write_position.height, min(write_position.extended_height, max(heights)))

        # Draw child panes.
        ypos = write_position.ypos
        xpos = write_position.xpos

        for s, c in zip(sizes, self.children):
            c.write_to_screen(cli, screen, WritePosition(xpos, ypos, s, height))
            xpos += s


class FloatContainer(Layout):
    """
    Container which can contain another container for the background, as well
    as a list of floating containers on top of it.

    Example Usage::

        FloatContainer(content=Window(...),
                       floats=[
                           Float(xcursor=True,
                                ycursor=True,
                                layout=CompletionMenu(...))
                       ])
    """
    def __init__(self, content, floats):
        assert isinstance(content, Layout)
        assert all(isinstance(f, Float) for f in floats)

        self.content = content
        self.floats = floats

    def reset(self):
        self.content.reset()

    def width(self, cli):
        return self.content.width(cli)

    def height(self, cli, width):
        """
        Return the preferred height of the float container.
        (We don't care about the height of the floats, they should always fit
        into the dimensions provided by the container.)
        """
        return self.content.height(cli, width)

    def write_to_screen(self, cli, screen, write_position):
        self.content.write_to_screen(cli, screen, write_position)

        # When a menu_position was given, use this instead of the cursor
        # position. (These cursor positions are absolute, translate again
        # relative to the write_position.)
        cursor_position = screen.menu_position or screen.cursor_position
        cursor_position = Point(x=cursor_position.x - write_position.xpos,
                                y=cursor_position.y - write_position.ypos)

        for fl in self.floats:
            # Left & width given.
            if fl.left is not None and fl.width is not None:
                xpos = fl.left
                width = fl.width
            # Left & right given -> calculate width.
            elif fl.left is not None and fl.right is not None:
                xpos = fl.left
                width = write_position.width - fl.left - fl.right
            # Width & right given -> calculate left.
            elif fl.width is not None and fl.right is not None:
                xpos = write_position.width - fl.right - fl.width
                width = fl.width
            elif fl.xcursor:
                width = fl.width
                if width is None:
                    width = fl.content.width(cli).preferred

                xpos = cursor_position.x
                if xpos + width > write_position.width:
                    xpos = max(0, write_position.width - width)
            else:
                xpos = 0
                width = write_position.width

            # Top & height given.
            if fl.top is not None and fl.height is not None:
                ypos = fl.top
                height = fl.height
            # Top & bottom given -> calculate height.
            elif fl.top is not None and fl.bottom is not None:
                ypos = fl.top
                height = write_position.height - fl.top - fl.bottom
            # Height & bottom given -> calculate top.
            elif fl.height is not None and fl.bottom is not None:
                ypos = write_position.height - fl.height - fl.bottom
                height = fl.height
            # Near cursor
            elif fl.ycursor:
                ypos = cursor_position.y + 1

                height = fl.height
                if height is None:
                    height = fl.content.height(cli, width).preferred

                # Reduce height if not enough space. (We can use the
                # extended_height when the content requires it.)
                if height > write_position.extended_height - ypos:
                    height = write_position.extended_height - ypos
            else:
                ypos = 0
                height = write_position.height

            # Write float.
            if xpos >= 0 and ypos >= 0 and height > 0 and width > 0:
                wp = WritePosition(xpos=xpos + write_position.xpos,
                                   ypos=ypos + write_position.ypos,
                                   width=width, height=height)
                fl.content.write_to_screen(cli, screen, wp)


class Float(object):
    def __init__(self, top=None, right=None, bottom=None, left=None,
                 width=None, height=None,
                 xcursor=False, ycursor=False, content=None):
        assert isinstance(content, Layout)

        self.left = left
        self.right = right
        self.top = top
        self.bottom = bottom

        self.width = width
        self.height = height

        self.xcursor = xcursor
        self.ycursor = ycursor

        self.content = content


class Window(Layout):
    """
    Layout that holds a control.

    :param content: User interface control.
    :param width: `LayoutDimension` instance.
    :param height: `LayoutDimension` instance.
    :param get_width: callable which takes a `CommandLineInterface` and returns a `LayoutDimension`.
    :param get_height: callable which takes a `CommandLineInterface` and returns a `LayoutDimension`.
    :param filter: `Filter` which decides about the visibility.
    :param dont_extend_width: When `True`, don't take up more width then the
                              preferred width reported by the control.
    :param dont_extend_height: When `True`, don't take up more width then the
                               preferred height reported by the control.
    """
    def __init__(self, content, width=None, height=None, get_width=None,
                 get_height=None, filter=None, dont_extend_width=False,
                 dont_extend_height=False):
        assert isinstance(content, UIControl)
        assert width is None or isinstance(width, LayoutDimension)
        assert height is None or isinstance(height, LayoutDimension)
        assert get_width is None or callable(get_width)
        assert get_height is None or callable(get_height)
        assert width is None or get_width is None
        assert height is None or get_height is None
        assert filter is None or isinstance(filter, Filter)

        self.content = content
        self.filter = filter
        self.dont_extend_width = dont_extend_width
        self.dont_extend_height = dont_extend_height
        self._width = get_width or (lambda cli: width)
        self._height = get_height or (lambda cli: height)

        self.reset()

    def __repr__(self):
        return 'Window(content=%r)' % self.content

    def reset(self):
        self.content.reset()

        #: Vertical scrolling position of the main content.
        self.vertical_scroll = 0

    def _visible(self, cli):
        return self.filter is None or self.filter(cli)

    def width(self, cli):
        if self._visible(cli):
            width = self._width(cli) or LayoutDimension()
            preferred_width = self.content.preferred_width(cli)

            if preferred_width is None:
                return width
            else:
                # When 'dont_extend_width' has been given. Don't use more than
                # the preferred width of the control. (But also don't go below
                # the minimum.)
                if self.dont_extend_width:
                    max_width = max(width.min, min(preferred_width, width.max))
                else:
                    max_width = width.max
                return LayoutDimension(min=width.min, max=max_width, preferred=preferred_width)
        else:
            return LayoutDimension.exact(0)

    def height(self, cli, width):
        if self._visible(cli):
            height = self._height(cli) or LayoutDimension()
            preferred_height = self.content.preferred_height(cli, width)

            if preferred_height is None:
                return height
            else:
                # When 'dont_extend_height' has been given. Don't use more than
                # the preferred height of the control. (But also don't go below
                # the minimum.)
                if self.dont_extend_height:
                    max_height = max(height.min, min(preferred_height, height.max))
                else:
                    max_height = height.max
                return LayoutDimension(min=height.min, max=max_height, preferred=preferred_height)
        else:
            return LayoutDimension.exact(0)

    def write_to_screen(self, cli, screen, write_position):
                        # XXX: Show window too small messsage...

        # Only write when visible.
        if self._visible(cli):
            # Set position.
            temp_screen = self.content.create_screen(cli, write_position.width, write_position.height)
            self._scroll(temp_screen, write_position.height)
            self._copy(cli, temp_screen, screen, write_position)

    def _copy(self, cli, temp_screen, new_screen, write_position):
        """
        Copy characters from the temp screen that we got from the `UIControl`
        to the real screen.
        """
        xpos = write_position.xpos
        ypos = write_position.ypos
        width = write_position.width
        height = write_position.height

        columns = temp_screen.width

        temp_buffer = temp_screen._buffer
        new_buffer = new_screen._buffer
        temp_screen_height = temp_screen.current_height
        y = 0

        # Now copy the region we need to the real screen.
        for y in range(0, height):
            # We keep local row variables. (Don't look up the row in the dict
            # for each iteration of the nested loop.)
            new_row = new_buffer[y + ypos]

            if y >= temp_screen_height and y >= write_position.height:
                # Break out of for loop when we pass after the last row of the
                # temp screen. (We use the 'y' position for calculation of new
                # screen's height.)
                break
            else:
                temp_row = temp_buffer[y + self.vertical_scroll]
                for x in range(0, columns):
                    new_row[x + xpos] = temp_row[x]

        if self.content.has_focus(cli):
            new_screen.cursor_position = Point(y=temp_screen.cursor_position.y + ypos - self.vertical_scroll,
                                               x=temp_screen.cursor_position.x + xpos)

        if not new_screen.menu_position and temp_screen.menu_position:
            new_screen.menu_position = Point(y=temp_screen.menu_position.y + ypos - self.vertical_scroll,
                                             x=temp_screen.menu_position.x + xpos)

        # Update height of this screen.
        new_screen.current_height = max(new_screen.current_height, ypos + y + 1)

    def _scroll(self, temp_screen, height):
        """
        Scroll to make sure the cursor position is visible.
        """
        # Scroll back if we scrolled to much and there's still space at the top.
        if self.vertical_scroll > temp_screen.current_height - height:
            self.vertical_scroll = max(0, temp_screen.current_height - height)

        # Scroll up if cursor is before visible part.
        if self.vertical_scroll > temp_screen.cursor_position.y:
            self.vertical_scroll = temp_screen.cursor_position.y

        # Scroll down if cursor is after visible part.
        if self.vertical_scroll <= temp_screen.cursor_position.y - height:
            self.vertical_scroll = (temp_screen.cursor_position.y + 1) - height



from __future__ import unicode_literals

from pygments.token import Token
from ..enums import IncrementalSearchDirection
from .utils import token_list_len
from .processors import Processor

__all__ = (
    'DefaultPrompt',
)


class DefaultPrompt(Processor):
    """
    Default prompt. This one shows the 'arg' and reverse search like
    Bash/readline normally do.
    """
    def __init__(self, prompt='> '):
        self.prompt = prompt

    def run(self, cli, buffer, tokens):
        # Get text before cursor.
        if buffer.isearch_state:
            before = _get_isearch_tokens(buffer.isearch_state)

        elif cli.input_processor.arg is not None:
            before = _get_arg_tokens(cli)

        else:
            before = [(Token.Prompt, self.prompt)]

        # Insert before buffer text.
        shift_position = token_list_len(before)

        return before + tokens, lambda i: i + shift_position

    def invalidation_hash(self, cli, buffer):
        return (
            cli.input_processor.arg,
            buffer.isearch_state,
            buffer.isearch_state and buffer.isearch_state.isearch_text,
        )


def _get_isearch_tokens(isearch_state):
    def before():
        if isearch_state.isearch_direction == IncrementalSearchDirection.BACKWARD:
            text = 'reverse-i-search'
        else:
            text = 'i-search'

        return [(Token.Prompt.ISearch, '(%s)`' % text)]

    def text():
        index = isearch_state.no_match_from_index
        text = isearch_state.isearch_text

        if index is None:
            return [(Token.Prompt.ISearch.Text, text)]
        else:
            return [
                (Token.Prompt.ISearch.Text, text[:index]),
                (Token.Prompt.ISearch.Text.NoMatch, text[index:])
            ]

    def after():
        return [(Token.Prompt.ISearch, '`: ')]

    return before() + text() + after()


def _get_arg_tokens(cli):
    """
    Tokens for the arg-prompt.
    """
    arg = cli.input_processor.arg

    return [
        (Token.Prompt.Arg, '(arg: '),
        (Token.Prompt.Arg.Text, str(arg)),
        (Token.Prompt.Arg, ') '),
    ]



from __future__ import unicode_literals
from pygments.formatters.terminal256 import Terminal256Formatter, EscapeSequence

from prompt_toolkit.layout.screen import Size

import array
import fcntl
import six
import termios
import errno


# Global variable to keep the colour table in memory.
_tf = Terminal256Formatter()

#: If True: write the output of the renderer also to the following file. This
#: is very useful for debugging. (e.g.: to see that we don't write more bytes
#: than required.)
_DEBUG_RENDER_OUTPUT = False
_DEBUG_RENDER_OUTPUT_FILENAME = '/tmp/prompt-toolkit-render-output'


def _get_size(fileno):
    # Thanks to fabric (fabfile.org), and
    # http://sqizit.bartletts.id.au/2011/02/14/pseudo-terminals-in-python/
    """
    Get the size of this pseudo terminal.

    :param fileno: stdout.fileno()
    :returns: A (rows, cols) tuple.
    """
#    assert stdout.isatty()

    # Buffer for the C call
    buf = array.array(u'h' if six.PY3 else b'h', [0, 0, 0, 0])

    # Do TIOCGWINSZ (Get)
    fcntl.ioctl(fileno, termios.TIOCGWINSZ, buf, True)

    # Return rows, cols
    return buf[0], buf[1]


class Vt100_Output(object):
    def __init__(self, stdout):
        self._buffer = []
        self.stdout = stdout

    def get_size(self):
        rows, columns = _get_size(self.stdout.fileno())
        return Size(rows=rows, columns=columns)

    def write(self, data):
        self._buffer.append(data)

    def erase_screen(self):
        """
        Erases the screen with the background colour and moves the cursor to
        home.
        """
        self.write('\x1b[2J')

    def enter_alternate_screen(self):
        self.write('\x1b[?1049h\x1b[H')

    def quit_alternate_screen(self):
        self.write('\x1b[?1049l')

    def erase_end_of_line(self):
        """
        Erases from the current cursor position to the end of the current line.
        """
        self.write('\x1b[K')

    def erase_down(self):
        """
        Erases the screen from the current line down to the bottom of the
        screen.
        """
        self.write('\x1b[J')

    def newline(self):
        self.write('\n')

    def carriage_return(self):
        self.write('\r')

    def reset_attributes(self):
        self.write('\x1b[0m')

    def set_attributes(self, fgcolor=None, bgcolor=None, bold=False, underline=False):
        """
        Create new style and output.
        """
        fg = _tf._color_index(fgcolor) if fgcolor else None
        bg = _tf._color_index(bgcolor) if bgcolor else None

        e = EscapeSequence(fg=fg, bg=bg, bold=bold, underline=underline)

        self.reset_attributes()
        self.write(e.color_string())

    def disable_autowrap(self):
        self.write('\x1b[?7l')

    def enable_autowrap(self):
        self.write('\x1b[?7h')

    def cursor_goto(self, row=0, column=0):
        """ Move cursor position. """
        self.write('\x1b[%i;%iH' % (row, column))

    def cursor_up(self, amount):
        if amount == 0:
            self.write('')
        elif amount == 1:
            self.write('\x1b[A')
        else:
            self.write('\x1b[%iA' % amount)

    def cursor_down(self, amount):
        if amount == 0:
            self.write('')
        elif amount == 1:
            # Note: Not the same as '\n', '\n' can cause the window content to
            #       scroll.
            self.write('\x1b[B')
        else:
            self.write('\x1b[%iB' % amount)

    def cursor_forward(self, amount):
        if amount == 0:
            self.write('')
        elif amount == 1:
            self.write('\x1b[C')
        else:
            self.write('\x1b[%iC' % amount)

    def cursor_backward(self, amount):
        if amount == 0:
            self.write('')
        elif amount == 1:
            self.write('\b')  # '\x1b[D'
        else:
            self.write('\x1b[%iD' % amount)

    def flush(self):
        """
        Write to output stream and flush.
        """
        if not self._buffer:
            return

        data = ''.join(self._buffer)

        try:
            self.stdout.write(data)
            self.stdout.flush()
        except IOError as e:
            if e.args and e.args[0] == errno.EINTR:
                # Interrupted system call. Can happpen in case of a window
                # resize signal. (Just ignore. The resize handler will render
                # again anyway.)
                pass
            else:
                raise

        self._buffer = []



from __future__ import unicode_literals

from prompt_toolkit.completion import Completer, Completion
import os

__all__ = (
    'PathCompleter',
)


class PathCompleter(Completer):
    """
    Complete for Path variables.

    :param get_paths: Callable which returns a list of directories to look into
                      when the user enters a relative path.
    :param file_filter: Callable which takes a filename and returns whether
                        this file should show up in the completion. ``None``
                        when no filtering has to be done.
    :param min_input_len: Don't do autocompletion when the input string is shorter.
    """
    def __init__(self, only_directories=False, get_paths=None, file_filter=None,
                 min_input_len=0):
        assert get_paths is None or callable(get_paths)
        assert file_filter is None or callable(file_filter)
        assert isinstance(min_input_len, int)

        self.only_directories = only_directories
        self.get_paths = get_paths or (lambda: ['.'])
        self.file_filter = file_filter or (lambda _: True)
        self.min_input_len = min_input_len

    def get_completions(self, document, complete_event):
        text = document.text_before_cursor

        # Complete only when we have at least the  minimal input length,
        # otherwise, we can too many results and autocompletion will become too
        # heavy.
        if len(text) < self.min_input_len:
            return

        try:
            dirname = os.path.dirname(text)
            directories = [os.path.dirname(text)] if dirname else self.get_paths()
            prefix = os.path.basename(text)

            # Get all filenames.
            filenames = []
            for directory in directories:
                for filename in os.listdir(directory):
                    if filename.startswith(prefix):
                        filenames.append((directory, filename))

            # Sort
            filenames = sorted(filenames, key=lambda k: k[1])

            # Yield them.
            for directory, filename in filenames:
                completion = filename[len(prefix):]
                full_name = os.path.join(directory, filename)

                if not os.path.isdir(full_name):
                    if self.only_directories or not self.file_filter(full_name):
                        continue

                yield Completion(completion, 0, display=filename,
                                 get_display_meta=self._create_meta_getter(full_name))
        except OSError:
            pass

    def _create_meta_getter(self, full_name):
        """
        Create lazy meta getter.
        """
        def getter():
            if os.path.isdir(full_name):
                return 'Directory'
            elif os.path.isfile(full_name):
                return 'File'
            elif os.path.islink(full_name):
                return 'Link'
            else:
                return ''
        return getter



# Generated: 2014-05-05T02:09:32.625406+00:00
# Source: DerivedCombiningClass-6.3.0.txt
# Date:  2013-07-05, 14:08:44 GMT [MD]
NONZERO_COMBINING = (
    (0x0300, 0x034e,),  # Combining Grave Accent  ..Combining Upwards Arrow
    (0x0350, 0x036f,),  # Combining Right Arrowhea..Combining Latin Small Le
    (0x0483, 0x0487,),  # Combining Cyrillic Titlo..Combining Cyrillic Pokry
    (0x0591, 0x05bd,),  # Hebrew Accent Etnahta   ..Hebrew Point Meteg
    (0x05bf, 0x05bf,),  # Hebrew Point Rafe       ..Hebrew Point Rafe
    (0x05c1, 0x05c2,),  # Hebrew Point Shin Dot   ..Hebrew Point Sin Dot
    (0x05c4, 0x05c5,),  # Hebrew Mark Upper Dot   ..Hebrew Mark Lower Dot
    (0x05c7, 0x05c7,),  # Hebrew Point Qamats Qata..Hebrew Point Qamats Qata
    (0x0610, 0x061a,),  # Arabic Sign Sallallahou ..Arabic Small Kasra
    (0x064b, 0x065f,),  # Arabic Fathatan         ..Arabic Wavy Hamza Below
    (0x0670, 0x0670,),  # Arabic Letter Superscrip..Arabic Letter Superscrip
    (0x06d6, 0x06dc,),  # Arabic Small High Ligatu..Arabic Small High Seen
    (0x06df, 0x06e4,),  # Arabic Small High Rounde..Arabic Small High Madda
    (0x06e7, 0x06e8,),  # Arabic Small High Yeh   ..Arabic Small High Noon
    (0x06ea, 0x06ed,),  # Arabic Empty Centre Low ..Arabic Small Low Meem
    (0x0711, 0x0711,),  # Syriac Letter Superscrip..Syriac Letter Superscrip
    (0x0730, 0x074a,),  # Syriac Pthaha Above     ..Syriac Barrekh
    (0x07eb, 0x07f3,),  # Nko Combining Short High..Nko Combining Double Dot
    (0x0816, 0x0819,),  # Samaritan Mark In       ..Samaritan Mark Dagesh
    (0x081b, 0x0823,),  # Samaritan Mark Epentheti..Samaritan Vowel Sign A
    (0x0825, 0x0827,),  # Samaritan Vowel Sign Sho..Samaritan Vowel Sign U
    (0x0829, 0x082d,),  # Samaritan Vowel Sign Lon..Samaritan Mark Nequdaa
    (0x0859, 0x085b,),  # Mandaic Affrication Mark..Mandaic Gemination Mark
    (0x08e4, 0x08fe,),  # Arabic Curly Fatha      ..Arabic Damma With Dot
    (0x093c, 0x093c,),  # Devanagari Sign Nukta   ..Devanagari Sign Nukta
    (0x094d, 0x094d,),  # Devanagari Sign Virama  ..Devanagari Sign Virama
    (0x0951, 0x0954,),  # Devanagari Stress Sign U..Devanagari Acute Accent
    (0x09bc, 0x09bc,),  # Bengali Sign Nukta      ..Bengali Sign Nukta
    (0x09cd, 0x09cd,),  # Bengali Sign Virama     ..Bengali Sign Virama
    (0x0a3c, 0x0a3c,),  # Gurmukhi Sign Nukta     ..Gurmukhi Sign Nukta
    (0x0a4d, 0x0a4d,),  # Gurmukhi Sign Virama    ..Gurmukhi Sign Virama
    (0x0abc, 0x0abc,),  # Gujarati Sign Nukta     ..Gujarati Sign Nukta
    (0x0acd, 0x0acd,),  # Gujarati Sign Virama    ..Gujarati Sign Virama
    (0x0b3c, 0x0b3c,),  # Oriya Sign Nukta        ..Oriya Sign Nukta
    (0x0b4d, 0x0b4d,),  # Oriya Sign Virama       ..Oriya Sign Virama
    (0x0bcd, 0x0bcd,),  # Tamil Sign Virama       ..Tamil Sign Virama
    (0x0c4d, 0x0c4d,),  # Telugu Sign Virama      ..Telugu Sign Virama
    (0x0c55, 0x0c56,),  # Telugu Length Mark      ..Telugu Ai Length Mark
    (0x0cbc, 0x0cbc,),  # Kannada Sign Nukta      ..Kannada Sign Nukta
    (0x0ccd, 0x0ccd,),  # Kannada Sign Virama     ..Kannada Sign Virama
    (0x0d4d, 0x0d4d,),  # Malayalam Sign Virama   ..Malayalam Sign Virama
    (0x0dca, 0x0dca,),  # Sinhala Sign Al-lakuna  ..Sinhala Sign Al-lakuna
    (0x0e38, 0x0e3a,),  # Thai Character Sara U   ..Thai Character Phinthu
    (0x0e48, 0x0e4b,),  # Thai Character Mai Ek   ..Thai Character Mai Chatt
    (0x0eb8, 0x0eb9,),  # Lao Vowel Sign U        ..Lao Vowel Sign Uu
    (0x0ec8, 0x0ecb,),  # Lao Tone Mai Ek         ..Lao Tone Mai Catawa
    (0x0f18, 0x0f19,),  # Tibetan Astrological Sig..Tibetan Astrological Sig
    (0x0f35, 0x0f35,),  # Tibetan Mark Ngas Bzung ..Tibetan Mark Ngas Bzung
    (0x0f37, 0x0f37,),  # Tibetan Mark Ngas Bzung ..Tibetan Mark Ngas Bzung
    (0x0f39, 0x0f39,),  # Tibetan Mark Tsa -phru  ..Tibetan Mark Tsa -phru
    (0x0f71, 0x0f72,),  # Tibetan Vowel Sign Aa   ..Tibetan Vowel Sign I
    (0x0f74, 0x0f74,),  # Tibetan Vowel Sign U    ..Tibetan Vowel Sign U
    (0x0f7a, 0x0f7d,),  # Tibetan Vowel Sign E    ..Tibetan Vowel Sign Oo
    (0x0f80, 0x0f80,),  # Tibetan Vowel Sign Rever..Tibetan Vowel Sign Rever
    (0x0f82, 0x0f84,),  # Tibetan Sign Nyi Zla Naa..Tibetan Mark Halanta
    (0x0f86, 0x0f87,),  # Tibetan Sign Lci Rtags  ..Tibetan Sign Yang Rtags
    (0x0fc6, 0x0fc6,),  # Tibetan Symbol Padma Gda..Tibetan Symbol Padma Gda
    (0x1037, 0x1037,),  # Myanmar Sign Dot Below  ..Myanmar Sign Dot Below
    (0x1039, 0x103a,),  # Myanmar Sign Virama     ..Myanmar Sign Asat
    (0x108d, 0x108d,),  # Myanmar Sign Shan Counci..Myanmar Sign Shan Counci
    (0x135d, 0x135f,),  # Ethiopic Combining Gemin..Ethiopic Combining Gemin
    (0x1714, 0x1714,),  # Tagalog Sign Virama     ..Tagalog Sign Virama
    (0x1734, 0x1734,),  # Hanunoo Sign Pamudpod   ..Hanunoo Sign Pamudpod
    (0x17d2, 0x17d2,),  # Khmer Sign Coeng        ..Khmer Sign Coeng
    (0x17dd, 0x17dd,),  # Khmer Sign Atthacan     ..Khmer Sign Atthacan
    (0x18a9, 0x18a9,),  # Mongolian Letter Ali Gal..Mongolian Letter Ali Gal
    (0x1939, 0x193b,),  # Limbu Sign Mukphreng    ..Limbu Sign Sa-i
    (0x1a17, 0x1a18,),  # Buginese Vowel Sign I   ..Buginese Vowel Sign U
    (0x1a60, 0x1a60,),  # Tai Tham Sign Sakot     ..Tai Tham Sign Sakot
    (0x1a75, 0x1a7c,),  # Tai Tham Sign Tone-1    ..Tai Tham Sign Khuen-lue
    (0x1a7f, 0x1a7f,),  # Tai Tham Combining Crypt..Tai Tham Combining Crypt
    (0x1b34, 0x1b34,),  # Balinese Sign Rerekan   ..Balinese Sign Rerekan
    (0x1b44, 0x1b44,),  # Balinese Adeg Adeg      ..Balinese Adeg Adeg
    (0x1b6b, 0x1b73,),  # Balinese Musical Symbol ..Balinese Musical Symbol
    (0x1baa, 0x1bab,),  # Sundanese Sign Pamaaeh  ..Sundanese Sign Virama
    (0x1be6, 0x1be6,),  # Batak Sign Tompi        ..Batak Sign Tompi
    (0x1bf2, 0x1bf3,),  # Batak Pangolat          ..Batak Panongonan
    (0x1c37, 0x1c37,),  # Lepcha Sign Nukta       ..Lepcha Sign Nukta
    (0x1cd0, 0x1cd2,),  # Vedic Tone Karshana     ..Vedic Tone Prenkha
    (0x1cd4, 0x1ce0,),  # Vedic Sign Yajurvedic Mi..Vedic Tone Rigvedic Kash
    (0x1ce2, 0x1ce8,),  # Vedic Sign Visarga Svari..Vedic Sign Visarga Anuda
    (0x1ced, 0x1ced,),  # Vedic Sign Tiryak       ..Vedic Sign Tiryak
    (0x1cf4, 0x1cf4,),  # Vedic Tone Candra Above ..Vedic Tone Candra Above
    (0x1dc0, 0x1de6,),  # Combining Dotted Grave A..Combining Latin Small Le
    (0x1dfc, 0x1dff,),  # Combining Double Inverte..Combining Right Arrowhea
    (0x20d0, 0x20dc,),  # Combining Left Harpoon A..Combining Four Dots Abov
    (0x20e1, 0x20e1,),  # Combining Left Right Arr..Combining Left Right Arr
    (0x20e5, 0x20f0,),  # Combining Reverse Solidu..Combining Asterisk Above
    (0x2cef, 0x2cf1,),  # Coptic Combining Ni Abov..Coptic Combining Spiritu
    (0x2d7f, 0x2d7f,),  # Tifinagh Consonant Joine..Tifinagh Consonant Joine
    (0x2de0, 0x2dff,),  # Combining Cyrillic Lette..Combining Cyrillic Lette
    (0x302a, 0x302f,),  # Ideographic Level Tone M..Hangul Double Dot Tone M
    (0x3099, 0x309a,),  # Combining Katakana-hirag..Combining Katakana-hirag
    (0xa66f, 0xa66f,),  # Combining Cyrillic Vzmet..Combining Cyrillic Vzmet
    (0xa674, 0xa67d,),  # Combining Cyrillic Lette..Combining Cyrillic Payer
    (0xa69f, 0xa69f,),  # Combining Cyrillic Lette..Combining Cyrillic Lette
    (0xa6f0, 0xa6f1,),  # Bamum Combining Mark Koq..Bamum Combining Mark Tuk
    (0xa806, 0xa806,),  # Syloti Nagri Sign Hasant..Syloti Nagri Sign Hasant
    (0xa8c4, 0xa8c4,),  # Saurashtra Sign Virama  ..Saurashtra Sign Virama
    (0xa8e0, 0xa8f1,),  # Combining Devanagari Dig..Combining Devanagari Sig
    (0xa92b, 0xa92d,),  # Kayah Li Tone Plophu    ..Kayah Li Tone Calya Plop
    (0xa953, 0xa953,),  # Rejang Virama           ..Rejang Virama
    (0xa9b3, 0xa9b3,),  # Javanese Sign Cecak Telu..Javanese Sign Cecak Telu
    (0xa9c0, 0xa9c0,),  # Javanese Pangkon        ..Javanese Pangkon
    (0xaab0, 0xaab0,),  # Tai Viet Mai Kang       ..Tai Viet Mai Kang
    (0xaab2, 0xaab4,),  # Tai Viet Vowel I        ..Tai Viet Vowel U
    (0xaab7, 0xaab8,),  # Tai Viet Mai Khit       ..Tai Viet Vowel Ia
    (0xaabe, 0xaabf,),  # Tai Viet Vowel Am       ..Tai Viet Tone Mai Ek
    (0xaac1, 0xaac1,),  # Tai Viet Tone Mai Tho   ..Tai Viet Tone Mai Tho
    (0xaaf6, 0xaaf6,),  # Meetei Mayek Virama     ..Meetei Mayek Virama
    (0xabed, 0xabed,),  # Meetei Mayek Apun Iyek  ..Meetei Mayek Apun Iyek
    (0xfb1e, 0xfb1e,),  # Hebrew Point Judeo-spani..Hebrew Point Judeo-spani
    (0xfe20, 0xfe26,),  # Combining Ligature Left ..Combining Conjoining Mac
    (0x101fd, 0x101fd,),  # Phaistos Disc Sign Combi..Phaistos Disc Sign Combi
    (0x10a0d, 0x10a0d,),  # Kharoshthi Sign Double R..Kharoshthi Sign Double R
    (0x10a0f, 0x10a0f,),  # Kharoshthi Sign Visarga ..Kharoshthi Sign Visarga
    (0x10a38, 0x10a3a,),  # Kharoshthi Sign Bar Abov..Kharoshthi Sign Dot Belo
    (0x10a3f, 0x10a3f,),  # Kharoshthi Virama       ..Kharoshthi Virama
    (0x11046, 0x11046,),  # Brahmi Virama           ..Brahmi Virama
    (0x110b9, 0x110ba,),  # Kaithi Sign Virama      ..Kaithi Sign Nukta
    (0x11100, 0x11102,),  # Chakma Sign Candrabindu ..Chakma Sign Visarga
    (0x11133, 0x11134,),  # Chakma Virama           ..Chakma Maayyaa
    (0x111c0, 0x111c0,),  # Sharada Sign Virama     ..Sharada Sign Virama
    (0x116b6, 0x116b7,),  # Takri Sign Virama       ..Takri Sign Nukta
    (0x1d165, 0x1d169,),  # Musical Symbol Combining..Musical Symbol Combining
    (0x1d16d, 0x1d172,),  # Musical Symbol Combining..Musical Symbol Combining
    (0x1d17b, 0x1d182,),  # Musical Symbol Combining..Musical Symbol Combining
    (0x1d185, 0x1d18b,),  # Musical Symbol Combining..Musical Symbol Combining
    (0x1d1aa, 0x1d1ad,),  # Musical Symbol Combining..Musical Symbol Combining
    (0x1d242, 0x1d244,),  # Combining Greek Musical ..Combining Greek Musical
)



from __future__ import unicode_literals

from pygments.token import Token
from prompt_toolkit.filters import HasCompletions, IsDone, AlwaysOn
from .controls import UIControl
from .containers import Window
from .dimension import LayoutDimension
from .screen import Screen

__all__ = (
    'CompletionsMenu',
)


class CompletionsMenuControl(UIControl):
    """
    Helper for drawing the complete menu to the screen.
    """
    def __init__(self):
        self.token = Token.Menu.Completions

    def has_focus(self, cli):
        return False

    def preferred_width(self, cli):
        complete_state = cli.current_buffer.complete_state
        if complete_state:
            menu_width = self._get_menu_width(500, complete_state)
            menu_meta_width = self._get_menu_meta_width(500, complete_state)

            return menu_width + menu_meta_width + 1
        else:
            return 0

    def preferred_height(self, cli, width):
        complete_state = cli.current_buffer.complete_state
        if complete_state:
            return len(complete_state.current_completions)
        else:
            return 0

    def create_screen(self, cli, width, height):
        """
        Write the menu to the screen object.
        """
        screen = Screen(width)

        complete_state = cli.current_buffer.complete_state
        if complete_state:
            completions = complete_state.current_completions
            index = complete_state.complete_index  # Can be None!

            # Calculate width of completions menu.
            menu_width = self._get_menu_width(width - 1, complete_state)
            menu_meta_width = self._get_menu_meta_width(width - 1 - menu_width, complete_state)
            show_meta = self._show_meta(complete_state)

            if menu_width + menu_meta_width + 1 < width:
                menu_width += width - (menu_width + menu_meta_width + 1)

            # Decide which slice of completions to show.
            if len(completions) > height and (index or 0) > height / 2:
                slice_from = min(
                    (index or 0) - height // 2,  # In the middle.
                    len(completions) - height  # At the bottom.
                )
            else:
                slice_from = 0

            slice_to = min(slice_from + height, len(completions))

            # Create a function which decides at which positions the scroll button should be shown.
            def is_scroll_button(row):
                items_per_row = float(len(completions)) / min(len(completions), height)
                items_on_this_row_from = row * items_per_row
                items_on_this_row_to = (row + 1) * items_per_row
                return items_on_this_row_from <= (index or 0) < items_on_this_row_to

            # Write completions to screen.
            tokens = []

            for i, c in enumerate(completions[slice_from:slice_to]):
                is_current_completion = (i + slice_from == index)

                if is_scroll_button(i):
                    button_token = self.token.ProgressButton
                else:
                    button_token = self.token.ProgressBar

                if tokens:
                    tokens += [(Token, '\n')]
                tokens += (self._get_menu_item_tokens(c, is_current_completion, menu_width) +
                           (self._get_menu_item_meta_tokens(c, is_current_completion, menu_meta_width)
                               if show_meta else []) +
                           [(button_token, ' '), ])

            screen.write_at_position(tokens, width)

        return screen

    def _show_meta(self, complete_state):
        """
        Return ``True`` if we need to show a column with meta information.
        """
        return any(c.display_meta for c in complete_state.current_completions)

    def _get_menu_width(self, max_width, complete_state):
        """
        Return the width of the main column.
        """
        return min(max_width, max(len(c.display) for c in complete_state.current_completions) + 2)

    def _get_menu_meta_width(self, max_width, complete_state):
        """
        Return the width of the meta column.
        """
        if self._show_meta(complete_state):
            return min(max_width, max(len(c.display_meta) for c in complete_state.current_completions) + 2)
        else:
            return 0

    def _get_menu_item_tokens(self, completion, is_current_completion, width):
        if is_current_completion:
            token = self.token.Completion.Current
        else:
            token = self.token.Completion

        text = self._trim_text(completion.display, width - 2)
        return [(token, ' %%-%is ' % (width - 2) % text)]

    def _get_menu_item_meta_tokens(self, completion, is_current_completion, width):
        if is_current_completion:
            token = self.token.Meta.Current
        else:
            token = self.token.Meta

        text = self._trim_text(completion.display_meta, width - 2)
        return [(token, ' %%-%is ' % (width - 2) % text or 'none')]

    def _trim_text(self, text, max_width):
        """
        Trim the text to `max_width`, append dots when the text is too long.
        """
        # TODO: support for double width characters.
        if len(text) > max_width:
            return (text[:max(1, max_width-3)] + '...')[:max_width]
        else:
            return text


class CompletionsMenu(Window):
    def __init__(self, max_height=None, extra_filter=AlwaysOn()):
        super(CompletionsMenu, self).__init__(
            content=CompletionsMenuControl(),
            width=LayoutDimension(min=8),
            height=LayoutDimension(min=1, max=max_height),
            # Show when there are completions but not at the point we are
            # returning the input.
            filter=HasCompletions() & ~IsDone() & extra_filter)



from __future__ import unicode_literals

from six import string_types
from prompt_toolkit.completion import Completer, Completion

__all__ = (
    'WordCompleter',
)


class WordCompleter(Completer):
    """
    Simple autocompletion on a list of words.

    :param words: List of words.
    :param ignore_case: If True, case-insensitive completion.
    :param meta_dict: Optional dict mapping words to their meta-information.
    """
    def __init__(self, words, ignore_case=False, meta_dict=None):
        self.words = list(words)
        self.ignore_case = ignore_case
        self.meta_dict = meta_dict or {}
        assert all(isinstance(w, string_types) for w in self.words)

        if ignore_case:
            self.words = [w.lower() for w in self.words]

    def get_completions(self, document, complete_event):
        word_before_cursor = document.get_word_before_cursor()

        if self.ignore_case:
            word_before_cursor = word_before_cursor.lower()

        for a in self.words:
            if a.startswith(word_before_cursor):
                display_meta = self.meta_dict.get(a, '')
                yield Completion(a, -len(word_before_cursor), display_meta=display_meta)



"""
User interface Controls for the layout.
"""
from __future__ import unicode_literals
from pygments.token import Token

from six import with_metaclass
from abc import ABCMeta, abstractmethod

from prompt_toolkit.filters import AlwaysOff, Filter
from .processors import Processor
from .screen import Screen, Char, Point
from .utils import token_list_width

__all__ = (
    'TokenListControl',
    'FillControl',
    'BufferControl',
)


class UIControl(with_metaclass(ABCMeta, object)):
    """
    Base class for all user interface controls.
    """
    def reset(self):
        # Default reset. (Doesn't have to be implemented.)
        pass

    def preferred_width(self, cli):
        return None

    def preferred_height(self, cli, width):
        return None

    def has_focus(self, cli):
        return False

    @abstractmethod
    def create_screen(self, cli, width, height):
        """
        Write the content at this position to the screen.
        """
        pass


class _SimpleLRUCache(object):
    """
    Very simple LRU cache.

    :param maxsize: Maximum size of the cache. (Don't make it too big.)
    """
    def __init__(self, maxsize=8):
        self.maxsize = maxsize
        self._cache = []  # List of (key, value).

    def get(self, key, getter_func):
        """
        Get object from the cache.
        If not found, call `getter_func` to resolve it, and put that on the top
        of the cache instead.
        """
        # Look in cache first.
        for k, v in self._cache:
            if k == key:
                return v

        # Not found? Get it.
        value = getter_func()
        self._cache.append((key, value))

        if len(self._cache) > self.maxsize:
            self._cache = self._cache[-self.maxsize:]

        return value


class TokenListControl(UIControl):
    def __init__(self, get_tokens, default_char=None, align_right=False):
        assert default_char is None or isinstance(default_char, Char)
        self.get_tokens = get_tokens
        self.default_char = default_char
        self.align_right = align_right

    def __repr__(self):
        return 'TokenListControl(%r)' % self.get_tokens

    def preferred_width(self, cli):
        return token_list_width(self.get_tokens(cli))

    def preferred_height(self, cli, width):
        screen = self.create_screen(cli, width, None)
        return screen.current_height

    def create_screen(self, cli, width, height):
        screen = Screen(width, self.default_char)

        # Get tokens
        tokens = self.get_tokens(cli)

        # Align right
        if self.align_right:
            used_width = token_list_width(tokens)
            available_width = width - used_width
            tokens = [(self.default_char.token, ' ' * available_width)] + tokens

        screen.write_at_position(tokens, width)
        return screen

    @classmethod
    def static(cls, tokens):
        def get_static_tokens(cli):
            return tokens
        return cls(get_static_tokens)


class FillControl(UIControl):
    """
    Fill whole control with characters with this token.
    (Also helpful for debugging.)
    """
    def __init__(self, character=' ', token=Token):
        self.token = token
        self.character = character

    def __repr__(self):
        return 'FillControl(character=%r, token=%r)' % (self.character, self.token)

    def reset(self):
        pass

    def has_focus(self, cli):
        return False

    def create_screen(self, cli, width, height):
        char = Char(self.character, self.token)
        screen = Screen(width, char)
        screen.current_height = height
        return screen


class BufferControl(UIControl):
    def __init__(self,
                 input_processors=None,
                 lexer=None,
                 show_line_numbers=AlwaysOff(),
                 buffer_name='default',
                 default_token=Token,
                 menu_position=None):
        assert input_processors is None or all(isinstance(i, Processor) for i in input_processors)
        assert isinstance(show_line_numbers, Filter)
        assert menu_position is None or callable(menu_position)

        self.input_processors = input_processors or []
        self.show_line_numbers = show_line_numbers
        self.buffer_name = buffer_name
        self.default_token = default_token
        self.menu_position = menu_position

        if lexer:
            self.lexer = lexer(
                stripnl=False,
                stripall=False,
                ensurenl=False)
        else:
            self.lexer = None

        #: LRU cache for the lexer.
        #: Often, due to cursor movement, undo/redo and window resizing
        #: operations, it happens that a short time, the same document has to be
        #: lexed. This is a faily easy way to cache such an expensive operation.
        self._token_lru_cache = _SimpleLRUCache(maxsize=8)

        #: Keep a similar cache for rendered screens. (when we scroll up/down
        #: through the screen, or when we change another buffer, we don't want
        #: to recreate the same screen again.)
        self._screen_lru_cache = _SimpleLRUCache(maxsize=8)

    def _buffer(self, cli):
        """
        The buffer object that contains the 'main' content.
        """
        return cli.buffers[self.buffer_name]

    def has_focus(self, cli):
        return cli.focus_stack.current == self.buffer_name

    def preferred_width(self, cli):
        # Return the length of the longest line.
        return max(map(len, self._buffer(cli).document.lines))

    def preferred_height(self, cli, width):
        # Draw content on a screen using this width. Measure the height of the
        # result.
        screen = self.create_screen(cli, width, None)
        return screen.current_height

    def _get_input_tokens(self, cli, buffer):
        """
        Tokenize input text for highlighting.
        Return (tokens, cursor_transform_functions) tuple.
        """
        def get():
            if self.lexer:
                tokens = list(self.lexer.get_tokens(buffer.text))
            else:
                tokens = [(self.default_token, buffer.text)]

            # Run all processors over the input.
            # (They can transform both the tokens and the cursor position.)
            cursor_transform_functions = []

            for p in self.input_processors:
                tokens, f = p.run(cli, buffer, tokens)
                cursor_transform_functions.append(f)

            return tokens, cursor_transform_functions

        key = (
            buffer.text,

            # When the search state changes, highlighting will be different.
            # TODO: maybe use a `Processor` for the highlighting!
            buffer.isearch_state,
            buffer.isearch_state and buffer.isearch_state.isearch_text,

            # Include invalidation_hashes from all processors.
            tuple(p.invalidation_hash(cli, buffer) for p in self.input_processors),
        )

        return self._token_lru_cache.get(key, get)

    def _get_highlighted_characters(self, buffer):
        """
        Return a dictionary that maps the index of input string characters to
        their Token in case of highlighting.
        """
        highlighted_characters = {}

        # In case of incremental search, highlight all matches.
        if buffer.isearch_state:
            for index in buffer.document.find_all(buffer.isearch_state.isearch_text):
                if index == buffer.cursor_position:
                    token = Token.SearchMatch.Current
                else:
                    token = Token.SearchMatch

                highlighted_characters.update(dict([
                    (x, token) for x in range(index, index + len(buffer.isearch_state.isearch_text))
                ]))

        # In case of selection, highlight all matches.
        selection_range = buffer.document.selection_range()
        if selection_range:
            from_, to = selection_range

            for i in range(from_, to):
                highlighted_characters[i] = Token.SelectedText

        return highlighted_characters

    def _margin(self, cli, buffer):
        """
        Return a function that fills in the margin.
        """
        decimals = max(3, len('%s' % buffer.document.line_count))

        def numberred_margin(line_number):
            if line_number is not None:
                return [(Token.LineNumber, u'%%%si ' % decimals % (line_number + 1))]
            else:
                return [(Token.LineNumber, ' ' * (decimals + 1))]

        def no_margin(line_number):
            return []

        if self.show_line_numbers(cli):
            return numberred_margin
        else:
            return no_margin

    def create_screen(self, cli, width, height):
        buffer = self._buffer(cli)

        def _create_screen():
            screen = Screen(width)

            # Get tokens
            # Note: we add the space character at the end, because that's where
            #       the cursor can also be.
            input_tokens, cursor_transform_functions = self._get_input_tokens(cli, buffer)
            input_tokens += [(Token, ' ')]

            # 'Explode' tokens in characters.
            input_tokens = [(token, c) for token, text in input_tokens for c in text]

            # Apply highlighting.
            # XXX: not correct -> should go before input transforms!!!
            if not (cli.is_exiting or cli.is_aborting or cli.is_returning):
                highlighted_characters = self._get_highlighted_characters(buffer)

                for index, token in highlighted_characters.items():
                    input_tokens[index] = (token, input_tokens[index][1])

            indexes_to_pos = screen.write_at_position(
                input_tokens,
                screen.width,
                margin=self._margin(cli, buffer))

            def cursor_position_to_xy(cursor_position):
                # First get the real token position by applying all
                # transformations from the input processors.
                for f in cursor_transform_functions:
                    cursor_position = f(cursor_position)

                # Then look up into the table.
                try:
                    return indexes_to_pos[cursor_position]
                except KeyError:
                    # This can fail with KeyError, but only if one of the
                    # processors is returning invalid key locations.
                    raise
                    # return 0, 0

            return screen, cursor_position_to_xy

        # Build a key for the caching. If any of these parameters changes, we
        # have to recreate a new screen.
        key = (
            # When the text changes, we obviously have to recreate a new screen.
            buffer.text,

            # When the width changes, line wrapping will be different.
            # TODO: allow to disable line wrapping. + in that case, remove 'width'
            width,

            # When line numbers are enabled/disabled.
            self.show_line_numbers(cli),

            # When the search state changes, highlighting will be different.
            # TODO: maybe use a `Processor` for the highlighting!
            buffer.isearch_state,
            buffer.isearch_state and buffer.isearch_state.isearch_text,

            # When the selection changes, and the selection is highlighted.
            buffer.document.selection_range(),  # XXX: also use a processor for that one.

            # Include invalidation_hashes from all processors.
            tuple(p.invalidation_hash(cli, buffer) for p in self.input_processors),
        )

        # Get from cache, or create if this doesn't exist yet.
        screen, cursor_position_to_xy = self._screen_lru_cache.get(key, _create_screen)

        x, y = cursor_position_to_xy(buffer.cursor_position)
        screen.cursor_position = Point(y=y, x=x)

        # If there is an auto completion going on, use that start point for a
        # pop-up menu position. (But only when this buffer has the focus --
        # there is only one place for a menu, determined by the focussed buffer.)
        if cli.current_buffer_name == self.buffer_name:
            menu_position = self.menu_position(cli) if self.menu_position else None
            if menu_position is not None:
                assert isinstance(menu_position, int)
                x, y = cursor_position_to_xy(menu_position)
                screen.menu_position = Point(y=y, x=x)
            elif buffer.complete_state:
                x, y = cursor_position_to_xy(buffer.complete_state.original_document.cursor_position)
                screen.menu_position = Point(y=y, x=x)
            else:
                screen.menu_position = None

        return screen



"""
`GrammarLexer` is compatible with Pygments lexers and can be used to highlight
the input using a regular grammar with token annotations.
"""
from __future__ import unicode_literals
from pygments.token import Token

from .compiler import _CompiledGrammar

__all__ = (
    'GrammarLexer',
)


class GrammarLexer(object):
    """
    Lexer which can be used for highlighting of tokens according to variables in the grammar.

    (It does not actual lexing of the string, but it exposes an API, compatible
    with the Pygments lexer class.)

    :param compiled_grammar: Grammar as returned by the `compile()` function.
    :param tokens: (optionally) Dictionary mapping variable names of the
                   regular grammar to the Pygments Token that should be used
                   for this part.
    :param lexers: (optionally) Dictionary mapping variable names of the
                   regular grammar to the lexers that should be used for this part.
                   (This can call other lexer classes recursively.)
    """
    def __init__(self, compiled_grammar, tokens=None, lexers=None):
        assert isinstance(compiled_grammar, _CompiledGrammar)
        assert lexers is None or isinstance(lexers, dict)
        assert tokens is None or isinstance(tokens, dict)

        self.compiled_grammar = compiled_grammar
        self.tokens = tokens
        self.lexers = dict((name, lexer(stripnl=False, stripall=False, ensurenl=False))
                           for name, lexer in (lexers or {}).items())

    def __call__(self, stripnl=False, stripall=False, ensurenl=False):
        """
        For compatibility with Pygments lexers.
        (Signature of Pygments Lexer.__init__)
        """
        return self

    def get_tokens(self, text):
        m = self.compiled_grammar.match_prefix(text)

        if m:
            characters = [[Token, c] for c in text]

            for v in m.variables():
                # If we have a Pygmenst lexer for this part of the input.
                # Tokenize recursively and apply tokens.
                lexer = self.lexers.get(v.varname)
                token = self.tokens.get(v.varname)

                if lexer:
                    lexer_tokens = lexer.get_tokens(text[v.start:v.stop])
                    i = v.start
                    for t, s in lexer_tokens:
                        for c in s:
                            if characters[i][0] == Token:
                                characters[i][0] = t
                            i += 1

                elif token:
                    for i in range(v.start, v.stop):
                        if characters[i][0] == Token:
                            characters[i][0] = token
            return characters
        else:
            return [(Token, text)]



"""
Clipboard for command line interface.
"""
from __future__ import unicode_literals
import six

from .selection import SelectionType

__all__ = (
    'Clipboard',
    'ClipboardData',
)


class ClipboardData(object):
    """
    Text on the clipboard.

    :param text: string
    :param type: :class:`~.ClipboardDataType`
    """
    def __init__(self, text='', type=SelectionType.CHARACTERS):
        assert isinstance(text, six.string_types)
        assert type in (SelectionType.CHARACTERS, SelectionType.LINES)

        self.text = text
        self.type = type


class Clipboard(object):
    """
    Clipboard for command line interface.
    """
    def __init__(self):
        self._data = None

    def set_data(self, data):
        """
        Set data to the clipboard.

        :param data: :class:`~.ClipboardData` instance.
        """
        assert isinstance(data, ClipboardData)

        if data.text:
            self._data = data

    def set_text(self, text):
        assert isinstance(text, six.string_types)

        if text:
            self._data = ClipboardData(text)

    def get_data(self):
        return self._data or ClipboardData()



from __future__ import unicode_literals

from ..terminal.win32_input import ConsoleInputReader
from ..win32_types import SECURITY_ATTRIBUTES
from .base import BaseEventLoop

from ctypes import windll, pointer, c_long
from ctypes.wintypes import DWORD, BOOL


__all__ = (
    'Win32EventLoop',
)

WAIT_TIMEOUT = 0x00000102


class Win32EventLoop(BaseEventLoop):
    def __init__(self, input_processor, stdin):
        super(Win32EventLoop, self).__init__(input_processor, stdin)
        self._event = _create_event()
        self._console_input_reader = ConsoleInputReader()
        self._calls_from_executor = []

        # XXX: There is still one bug here. When input has been read from the
        #      ConsoleInputReader, `_wait_for_handles` never returns the Event
        #      as signalled anymore.

    def loop(self):
        if self.closed:
            raise Exception('Event loop already closed.')

        timeout = int(1000 * self.input_timeout)

        while True:
            handle = _wait_for_handles([self._event, self._console_input_reader.handle], timeout)

            if handle == self._event:
                windll.kernel32.ResetEvent(self._event)
                self._process_queued_calls_from_executor()
                return

            elif handle == self._console_input_reader.handle:
                keys = self._console_input_reader.read()
                for k in keys:
                    self.input_processor.feed_key(k)
                return

            else:
                # Fire input timeout event.
                self.onInputTimeout.fire()
                timeout = -1

    def close(self):
        super(Win32EventLoop, self).close()

        # Clean up Event object.
        windll.kernel32.CloseHandle(self._event)

    def call_from_executor(self, callback):
        """
        Call this function in the main event loop.
        Similar to Twisted's ``callFromThread``.
        """
        # Append to list of pending callbacks.
        self._calls_from_executor.append(callback)

        # Set Windows event.
        windll.kernel32.SetEvent(self._event)

    def _process_queued_calls_from_executor(self):
        # Process calls from executor.
        calls_from_executor, self._calls_from_executor = self._calls_from_executor, []
        for c in calls_from_executor:
            c()


def _wait_for_handles(handles, timeout=-1):
    """
    Waits for multiple handles. (Similar to 'select') Returns the handle which is ready.

    http://msdn.microsoft.com/en-us/library/windows/desktop/ms687025(v=vs.85).aspx
    """
    arrtype = c_long * len(handles)
    handle_array = arrtype(*handles)
    ret = windll.kernel32.WaitForMultipleObjects(len(handle_array), handle_array, False, DWORD(timeout))

    if ret == WAIT_TIMEOUT:
        return None
    else:
        h = handle_array[ret]
        return h


def _create_event():
    """
    Creates a Win32 unnamed Event .

    http://msdn.microsoft.com/en-us/library/windows/desktop/ms682396(v=vs.85).aspx
    """
    return windll.kernel32.CreateEventA(pointer(SECURITY_ATTRIBUTES()), BOOL(True), BOOL(False), None)



"""
Push/pop stack of buffer names. The top buffer of the stack is the one that
currently has the focus.
"""
from __future__ import unicode_literals
from six import string_types

__all__ = (
    'FocusStack',
)


class FocusStack(object):
    def __init__(self, initial='default'):
        self._initial = initial
        self.reset()

    def __repr__(self):
        return 'FocusStack(initial=%r, _stack=%r)' % (self._initial, self._stack)

    def reset(self):
        self._stack = [self._initial]

    def pop(self):
        if len(self._stack) > 1:
            self._stack.pop()
        else:
            raise Exception('Cannot pop last item from the focus stack.')

    def replace(self, buffer_name):
        assert isinstance(buffer_name, string_types)
        self._stack.pop()
        self._stack.append(buffer_name)

    def push(self, buffer_name):
        assert isinstance(buffer_name, string_types)
        self._stack.append(buffer_name)

    @property
    def current(self):
        return self._stack[-1]

    @property
    def previous(self):
        """
        Return the name of the previous focussed buffer, or return None.
        """
        if len(self._stack) > 1:
            return self._stack[-2]



from __future__ import unicode_literals
import os
import fcntl
import select
import signal
import errno

from codecs import getincrementaldecoder
from ..terminal.vt100_input import InputStream

from .base import BaseEventLoop

__all__ = (
    'PosixEventLoop',
    'call_on_sigwinch',
)


class PosixEventLoop(BaseEventLoop):
    stdin_decoder_cls = getincrementaldecoder('utf-8')

    def __init__(self, input_processor, stdin):
        super(PosixEventLoop, self).__init__(input_processor, stdin)

        self.inputstream = InputStream(self.input_processor)
        self._calls_from_executor = []

        # Create a pipe for inter thread communication.
        self._schedule_pipe = os.pipe()
        fcntl.fcntl(self._schedule_pipe[0], fcntl.F_SETFL, os.O_NONBLOCK)

        # Create incremental decoder for decoding stdin.
        # We can not just do `os.read(stdin.fileno(), 1024).decode('utf-8')`, because
        # it could be that we are in the middle of a utf-8 byte sequence.
        self._stdin_decoder = self.stdin_decoder_cls()

    def loop(self):
        """
        The input 'event loop'.
        """
        if self.closed:
            raise Exception('Event loop already closed.')

        timeout = self.input_timeout

        while True:
            r, w, x = _select([self.stdin, self._schedule_pipe[0]], [], [], timeout)

            # If we got a character, feed it to the input stream. If we got
            # none, it means we got a repaint request.
            if self.stdin in r:
                # Feed input text.
                data = self._read_from_stdin()
                self.inputstream.feed_and_flush(data)
                return

            # If we receive something on our "call_from_executor" pipe, process
            # these callbacks in a thread safe way.
            elif self._schedule_pipe[0] in r:
                # Flush all the pipe content.
                os.read(self._schedule_pipe[0], 1024)

                # Process calls from executor.
                calls_from_executor, self._calls_from_executor = self._calls_from_executor, []
                for c in calls_from_executor:
                    c()
            else:
                # Fire input timeout event.
                self.onInputTimeout.fire()
                timeout = None

    def _read_from_stdin(self):
        """
        Read the input and return it.
        """
        # Note: the following works better than wrapping `self.stdin` like
        #       `codecs.getreader('utf-8')(stdin)` and doing `read(1)`.
        #       Somehow that causes some latency when the escape
        #       character is pressed. (Especially on combination with the `select`.
        try:
            bytes = os.read(self.stdin.fileno(), 1024)
        except OSError:
            # In case of SIGWINCH
            bytes = b''

        try:
            return self._stdin_decoder.decode(bytes)
        except UnicodeDecodeError:
            # When it's not possible to decode this bytes, reset the decoder.
            # The only occurence of this that I had was when using iTerm2 on OS
            # X, with "Option as Meta" checked (You should choose "Option as
            # +Esc".)
            self._stdin_decoder = self.stdin_decoder_cls()
            return ''

    def call_from_executor(self, callback):
        """
        Call this function in the main event loop.
        Similar to Twisted's ``callFromThread``.
        """
        self._calls_from_executor.append(callback)

        if self._schedule_pipe:
            os.write(self._schedule_pipe[1], b'x')

    def close(self):
        super(PosixEventLoop, self).close()

        # Close pipes.
        schedule_pipe = self._schedule_pipe
        self._schedule_pipe = None

        if schedule_pipe:
            os.close(schedule_pipe[0])
            os.close(schedule_pipe[1])


def _select(*args, **kwargs):
    """
    Wrapper around select.select.

    When the SIGWINCH signal is handled, other system calls, like select
    are aborted in Python. This wrapper will retry the system call.
    """
    while True:
        try:
            return select.select(*args, **kwargs)
        except select.error as e:
            # Retry select call when EINTR
            if e.args and e.args[0] == errno.EINTR:
                continue
            else:
                raise


class call_on_sigwinch(object):
    """
    Context manager which Installs a SIGWINCH callback.
    (This signal occurs when the terminal size changes.)
    """
    def __init__(self, callback):
        self.callback = callback

    def __enter__(self):
        self.previous_callback = signal.signal(signal.SIGWINCH, lambda *a: self.callback())

    def __exit__(self, *a, **kw):
        signal.signal(signal.SIGWINCH, self.previous_callback)



from __future__ import unicode_literals

from prompt_toolkit.keys import Keys
import prompt_toolkit.filters as filters

from ..input_processor import KeyPress
from .utils import create_handle_decorator


__all__ = (
    'load_basic_bindings',
)


def load_basic_bindings(registry, filter=None):
    handle = create_handle_decorator(registry, filter)
    has_selection = filters.HasSelection()

    @handle(Keys.ControlA)
    @handle(Keys.ControlB)
    @handle(Keys.ControlC)
    @handle(Keys.ControlD)
    @handle(Keys.ControlE)
    @handle(Keys.ControlF)
    @handle(Keys.ControlG)
    @handle(Keys.ControlH)
    @handle(Keys.ControlI)
    @handle(Keys.ControlJ)
    @handle(Keys.ControlK)
    @handle(Keys.ControlL)
    @handle(Keys.ControlM)
    @handle(Keys.ControlN)
    @handle(Keys.ControlO)
    @handle(Keys.ControlP)
    @handle(Keys.ControlQ)
    @handle(Keys.ControlR)
    @handle(Keys.ControlS)
    @handle(Keys.ControlT)
    @handle(Keys.ControlU)
    @handle(Keys.ControlV)
    @handle(Keys.ControlW)
    @handle(Keys.ControlX)
    @handle(Keys.ControlY)
    @handle(Keys.ControlZ)
    @handle(Keys.F1)
    @handle(Keys.F2)
    @handle(Keys.F3)
    @handle(Keys.F4)
    @handle(Keys.F5)
    @handle(Keys.F6)
    @handle(Keys.F7)
    @handle(Keys.F8)
    @handle(Keys.F9)
    @handle(Keys.F10)
    @handle(Keys.F11)
    @handle(Keys.F12)
    @handle(Keys.F13)
    @handle(Keys.F14)
    @handle(Keys.F15)
    @handle(Keys.F16)
    @handle(Keys.F17)
    @handle(Keys.F18)
    @handle(Keys.F19)
    @handle(Keys.F20)
    @handle(Keys.ControlSpace)
    @handle(Keys.ControlBackslash)
    @handle(Keys.ControlSquareClose)
    @handle(Keys.ControlCircumflex)
    @handle(Keys.ControlUnderscore)
    @handle(Keys.Backspace)
    @handle(Keys.Up)
    @handle(Keys.Down)
    @handle(Keys.Right)
    @handle(Keys.Left)
    @handle(Keys.Home)
    @handle(Keys.End)
    @handle(Keys.Delete)
    @handle(Keys.ShiftDelete)
    @handle(Keys.PageUp)
    @handle(Keys.PageDown)
    @handle(Keys.BackTab)
    @handle(Keys.Tab)
    @handle(Keys.ControlLeft)
    @handle(Keys.ControlRight)
    @handle(Keys.ControlUp)
    @handle(Keys.ControlDown)
    def _(event):
        """
        First, for any of these keys, Don't do anything by default. Also don't
        catch them in the 'Any' handler which will insert them as data.

        If people want to insert these characters as a literal, they can always
        do by doing a quoted insert. (ControlQ in emacs mode, ControlV in Vi
        mode.)
        """
        pass

    @handle(Keys.Home)
    def _(event):
        buffer = event.current_buffer
        buffer.cursor_position += buffer.document.home_position

    @handle(Keys.End)
    def _(event):
        buffer = event.current_buffer
        buffer.cursor_position += buffer.document.end_position

    # CTRL keys.

    @handle(Keys.ControlC)
    def _(event):
        if event.current_buffer.returnable(event.cli):
            event.cli.set_abort()

    @handle(Keys.ControlD)  # XXX: this is emacs behaviour.
    def _(event):
        buffer = event.current_buffer

        # When there is text, act as delete, otherwise call exit.
        if buffer.text:
            buffer.delete()
        else:
            event.cli.set_exit()

    @handle(Keys.ControlI, filter= ~has_selection)
    def _(event):
        r"""
        Ctrl-I is identical to "\t"

        Traditional tab-completion, where the first tab completes the common
        suffix and the second tab lists all the completions.
        """
        buffer = event.current_buffer

        def second_tab():
            buffer.complete_next(start_at_first=False)

        # On the second tab-press, or when already navigating through
        # completions.
        if event.second_press or (buffer.complete_state and buffer.complete_state.complete_index is not None):
            second_tab()
        else:
            # On the first tab press, only complete the common parts of all completions.
            has_common = buffer.complete_common()
            if not has_common:
                second_tab()

    @handle(Keys.BackTab, filter= ~has_selection)
    def _(event):
        """
        Shift+Tab: go to previous completion.
        """
        event.current_buffer.complete_previous()

    @handle(Keys.ControlM)
    def _(event):
        """
        Transform ControlM (\r) by default into ControlJ (\n).
        (This way it is sufficient for other key bindings to handle only ControlJ.)
        Windows sends \r instead of \n when pressing enter.
        """
        def feed():
            event.cli.input_processor.feed_key(KeyPress(Keys.ControlJ, ''))

        # We use `call_from_executor`, to schedule this for later execution,
        # otherwise, we're sending data into a generator which is already
        # executing.
        event.cli.call_from_executor(feed)

    @handle(Keys.Escape, Keys.ControlM)
    def _(event):
        """
        Transform Esc-ControlM into Esc-ControlJ.
        """
        def feed():
            event.cli.input_processor.feed_key(KeyPress(Keys.Escape, ''))
            event.cli.input_processor.feed_key(KeyPress(Keys.ControlJ, ''))

        event.cli.call_from_executor(feed)

    @handle(Keys.ControlJ, filter= ~has_selection)
    def _(event):
        """
        Newline/Enter. (Or return input.)
        """
        buffer = event.current_buffer

        if buffer.is_multiline:
            buffer.newline()
        else:
            if event.current_buffer.returnable(event.cli) and buffer.validate():
                event.current_buffer.add_to_history()
                event.cli.set_return_value(buffer.document)

    @handle(Keys.ControlK, filter= ~has_selection)
    def _(event):
        buffer = event.current_buffer
        deleted = buffer.delete(count=buffer.document.get_end_of_line_position())
        event.cli.clipboard.set_text(deleted)

    @handle(Keys.ControlL)
    def _(event):
        event.cli.renderer.clear()

    @handle(Keys.ControlT, filter= ~has_selection)
    def _(event):
        event.current_buffer.swap_characters_before_cursor()

    @handle(Keys.ControlU, filter= ~has_selection)
    def _(event):
        """
        Clears the line before the cursor position. If you are at the end of
        the line, clears the entire line.
        """
        buffer = event.current_buffer
        deleted = buffer.delete_before_cursor(count=-buffer.document.get_start_of_line_position())
        event.cli.clipboard.set_text(deleted)

    @handle(Keys.ControlW, filter= ~has_selection)
    def _(event):
        """
        Delete the word before the cursor.
        """
        buffer = event.current_buffer
        pos = buffer.document.find_start_of_previous_word(count=event.arg)

        if pos:
            deleted = buffer.delete_before_cursor(count=-pos)
            event.cli.clipboard.set_text(deleted)

    @handle(Keys.PageUp, filter= ~has_selection)
    def _(event):
        event.current_buffer.history_backward()

    @handle(Keys.PageDown, filter= ~has_selection)
    def _(event):
        event.current_buffer.history_forward()

    @handle(Keys.Left)
    def _(event):
        buffer = event.current_buffer
        buffer.cursor_position += buffer.document.get_cursor_left_position(count=event.arg)

    @handle(Keys.Right)
    def _(event):
        buffer = event.current_buffer
        buffer.cursor_position += buffer.document.get_cursor_right_position(count=event.arg)

    @handle(Keys.Up, filter= ~has_selection)
    def _(event):
        event.current_buffer.auto_up(count=event.arg)

    @handle(Keys.Up, filter=has_selection)
    def _(event):
        event.current_buffer.cursor_up(count=event.arg)

    @handle(Keys.Down, filter= ~has_selection)
    def _(event):
        event.current_buffer.auto_down(count=event.arg)

    @handle(Keys.Down, filter=has_selection)
    def _(event):
        event.current_buffer.cursor_down(count=event.arg)

    @handle(Keys.ControlH, filter= ~has_selection)
    @handle(Keys.Backspace, filter= ~has_selection)
    def _(event):
        buffer = event.current_buffer
        buffer.delete_before_cursor(count=event.arg)

    @handle(Keys.Delete, filter= ~has_selection)
    def _(event):
        event.current_buffer.delete(count=event.arg)

    @handle(Keys.ShiftDelete, filter= ~has_selection)
    def _(event):
        event.current_buffer.delete(count=event.arg)

    @handle(Keys.Any, filter= ~has_selection)
    def _(event):
        """
        Insert data at cursor position.
        """
        event.current_buffer.insert_text(event.data * event.arg)

    @handle(Keys.CPRResponse)
    def _(event):
        """
        Handle incoming Cursor-Position-Request response.
        """
        # The incoming data looks like u'\x1b[35;1R'
        # Parse row/col information.
        row, col = map(int, event.data[2:-1].split(';'))

        # Report absolute cursor position to the renderer.
        event.cli.renderer.report_absolute_cursor_row(row)

    @handle(Keys.ControlZ)
    def _(event):
        """
        Suspend process to background.
        """
        event.cli.suspend_to_background()



"""
Useful shortcuts for creating a `CommandLineInterface` and reading input from it.
------------------------------------------------------------------------ --------

If you are using this library for retrieving some input from the user (as a
pure Python replacement for GNU readline), probably for 90% of the use cases,
the `get_input` function is all you need. It's the easiest shortcut which does
a lot of the underlying work like creating a `CommandLineInterface` instance
for you.

When is this not sufficient:
    - When you want to have more complicated layouts (maybe with sidebars or
      multiple toolbars. Or visibility of certain user interface controls
      according to some conditions.)
    - When you wish to have multiple input buffers. (If you would create an
      editor like a Vi clone.)
    - Something else that requires more customization than what is possible
      with the parameters of `get_input`.

In that case, study the code in this file and build your own
`CommandLineInterface` instance. It's not too complicated.
"""
from __future__ import unicode_literals

from prompt_toolkit import CommandLineInterface, AbortAction
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.filters import IsDone, HasFocus
from prompt_toolkit.history import History, FileHistory
from prompt_toolkit.key_binding.manager import KeyBindingManager
from prompt_toolkit.layout import Window, HSplit, FloatContainer, Float
from prompt_toolkit.layout.controls import BufferControl, TokenListControl
from prompt_toolkit.layout.dimension import LayoutDimension
from prompt_toolkit.layout.menus import CompletionsMenu
from prompt_toolkit.layout.processors import PasswordProcessor
from prompt_toolkit.layout.prompt import DefaultPrompt
from prompt_toolkit.layout.screen import Char
from prompt_toolkit.layout.toolbars import ValidationToolbar, SystemToolbar

from pygments.token import Token


__all__ = (
    'get_input',
    'create_cli',
    'create_default_layout',
)


def create_default_layout(message='', lexer=None, is_password=False,
                          reserve_space_for_menu=False, get_bottom_toolbar_tokens=None):
    """
    Generate default layout.
    """
    assert get_bottom_toolbar_tokens is None or callable(get_bottom_toolbar_tokens)

    # Create processors list.
    if is_password:
        input_processors = [PasswordProcessor(), DefaultPrompt(message)]
    else:
        input_processors = [DefaultPrompt(message)]

    # Create bottom toolbar.
    if get_bottom_toolbar_tokens:
        toolbars = [Window(TokenListControl(get_bottom_toolbar_tokens,
                                            default_char=Char(' ', Token.Toolbar)),
                           height=LayoutDimension.exact(1),
                           filter=~IsDone())]
    else:
        toolbars = []

    def get_height(cli):
        # If there is an autocompletion menu to be shown, make sure that our
        # layout has at least a minimal height in order to display it.
        if reserve_space_for_menu and not cli.is_done:
            return LayoutDimension(min=8)
        else:
            return LayoutDimension()

    # Create and return Layout instance.
    return HSplit([
        FloatContainer(
            Window(
                BufferControl(
                    input_processors=input_processors,
                    lexer=lexer),
                get_height=get_height,
            ),
            [
                Float(xcursor=True,
                      ycursor=True,
                      content=CompletionsMenu(max_height=16,
                                              extra_filter=HasFocus('default')))
            ]
        ),
        ValidationToolbar(),
        SystemToolbar(),
    ] + toolbars)


def create_cli(message='',
               multiline=False,
               is_password=False,
               vi_mode=False,
               lexer=None,
               enable_system_prompt=False,
               enable_open_in_editor=False,
               validator=None,
               completer=None,
               style=None,
               history_filename=None,
               get_bottom_toolbar_tokens=None):

    # Create history instance.
    if history_filename:
        history = FileHistory(history_filename)
    else:
        history = History()

    # Load all key bindings.
    manager = KeyBindingManager(enable_vi_mode=vi_mode,
                                enable_system_prompt=enable_system_prompt,
                                enable_open_in_editor=enable_open_in_editor)

    # Create interface.
    return CommandLineInterface(
        layout=create_default_layout(message=message, lexer=lexer, is_password=is_password,
                                     reserve_space_for_menu=(completer is not None),
                                     get_bottom_toolbar_tokens=get_bottom_toolbar_tokens),
        buffer=Buffer(
            is_multiline=multiline,
            history=history,
            validator=validator,
            completer=completer,
        ),
        key_bindings_registry=manager.registry,
        style=style)


def get_input(message='',
              raise_exception_on_abort=False,
              multiline=False,
              is_password=False,
              vi_mode=False,
              lexer=None,
              validator=None,
              completer=None,
              style=None,
              enable_system_prompt=False,
              enable_open_in_editor=False,
              history_filename=None,
              get_bottom_toolbar_tokens=None):
    """
    Get input from the user and return it. This wrapper builds the most obvious
    configuration of a `CommandLineInterface`. This can be a replacement for
    `raw_input`. (or GNU readline.)

    This returns `None` when Ctrl-D was pressed.

    :param message: Text to be shown before the prompt.
    :param mulitiline: Allow multiline input. Pressing enter will insert a
                       newline. (This requires Meta+Enter to accept the input.)
    :param is_password: Show asterisks instead of the actual typed characters.
    :param vi_mode: If True, use Vi key bindings.
    :param lexer: Pygments lexer to be used for the syntax highlighting.
    :param validator: `Validator` instance for input validation.
    :param completer: `Completer` instance for input completion.
    :param style: Pygments style class for the color scheme.
    :param enable_system_prompt: Pressing Meta+'!' will show a system prompt.
    :param enable_open_in_editor: Pressing 'v' in Vi mode or C-X C-E in emacs
                                  mode will open an external editor.
    :param history_filename: If not `None`, keep a persistent history in this file.
    :param get_bottom_toolbar_tokens: Optional callable which takes a
        :class:`CommandLineInterface` and returns a list of tokens for the
        bottom toolbar.
    """
    cli = create_cli(
        message=message,
        multiline=multiline,
        is_password=is_password,
        vi_mode=vi_mode,
        lexer=lexer,
        enable_system_prompt=enable_system_prompt,
        enable_open_in_editor=enable_open_in_editor,
        validator=validator,
        completer=completer,
        style=style,
        history_filename=history_filename,
        get_bottom_toolbar_tokens=get_bottom_toolbar_tokens)

    # Read input and return it.
    on_abort = AbortAction.RAISE_EXCEPTION if raise_exception_on_abort else AbortAction.RETURN_NONE
    document = cli.read_input(on_abort=on_abort, on_exit=AbortAction.IGNORE)

    if document:
        return document.text



"""
DEPRECATED.
"""
from __future__ import unicode_literals

__all__ = ('PythonRepl', 'embed')


# We have this for compatibility with older versions of:
# django_extensions.management.commands.shell_plus
try:
    import ptpython.repl

    def embed(*a, **kw):
        """
        DEPRECATED. Only for backwards compatibility.
        Please call ptpython.repl.embed directly!
        """
        ptpython.repl.embed(*a, **kw)
except ImportError as e:
    pass



from __future__ import unicode_literals
from prompt_toolkit.buffer import SelectionType, indent, unindent
from prompt_toolkit.keys import Keys
from prompt_toolkit.enums import IncrementalSearchDirection

from .basic import load_basic_bindings
from .utils import create_handle_decorator, focus_next_buffer

import prompt_toolkit.filters as filters

__all__ = (
    'load_emacs_bindings',
    'load_emacs_search_bindings',
    'load_emacs_system_bindings',
)


def load_emacs_bindings(registry, filter=None):
    """
    Some e-macs extensions.
    """
    # Overview of Readline emacs commands:
    # http://www.catonmat.net/download/readline-emacs-editing-mode-cheat-sheet.pdf

    load_basic_bindings(registry, filter)

    handle = create_handle_decorator(registry, filter)
    has_selection = filters.HasSelection()

    @handle(Keys.Escape)
    def _(event):
        """
        By default, ignore escape key.

        (If we don't put this here, and Esc is followed by a key which sequence
        is not handled, we'll insert an Escape character in the input stream.
        Something we don't want and happens to easily in emacs mode.
        Further, people can always use ControlQ to do a quoted insert.)
        """
        pass

    @handle(Keys.ControlA)
    def _(event):
        """
        Start of line.
        """
        buffer = event.current_buffer
        buffer.cursor_position += buffer.document.get_start_of_line_position(after_whitespace=False)

    @handle(Keys.ControlB)
    def _(event):
        """
        Character back.
        """
        buffer = event.current_buffer
        buffer.cursor_position += buffer.document.get_cursor_left_position(count=event.arg)

    @handle(Keys.ControlE)
    def _(event):
        """
        End of line.
        """
        buffer = event.current_buffer
        buffer.cursor_position += buffer.document.get_end_of_line_position()

    @handle(Keys.ControlF)
    def _(event):
        """
        Character forward.
        """
        buffer = event.current_buffer
        buffer.cursor_position += buffer.document.get_cursor_right_position(count=event.arg)

    @handle(Keys.ControlN, filter= ~has_selection)
    def _(event):
        """
        Next line.
        """
        event.current_buffer.auto_down()

    @handle(Keys.ControlN, filter=has_selection)
    def _(event):
        """
        Next line.
        """
        event.current_buffer.cursor_down()

    @handle(Keys.ControlO, filter= ~has_selection)
    def _(event):
        """
        Insert newline, but don't move the cursor.
        """
        event.current_buffer.insert_text('\n', move_cursor=False)

    @handle(Keys.ControlP, filter= ~has_selection)
    def _(event):
        """
        Previous line.
        """
        event.current_buffer.auto_up()

    @handle(Keys.ControlP, filter=has_selection)
    def _(event):
        """
        Previous line.
        """
        event.current_buffer.cursor_up()

    @handle(Keys.ControlQ, Keys.Any, filter= ~has_selection)
    def _(event):
        """
        Quoted insert.
        """
        event.current_buffer.insert_text(event.data, overwrite=False)

    @handle(Keys.ControlY, filter= ~has_selection)
    @handle(Keys.ControlX, 'r', 'y', filter= ~has_selection)
    def _(event):
        """
        Paste before cursor.
        """
        event.current_buffer.paste_clipboard_data(
            event.cli.clipboard.get_data(), count=event.arg, before=True)

    @handle(Keys.ControlUnderscore, save_before=False, filter= ~has_selection)
    def _(event):
        """
        Undo.
        """
        event.current_buffer.undo()

    def handle_digit(c):
        """
        Handle Alt + digit in the `meta_digit` method.
        """
        @handle(Keys.Escape, c)
        def _(event):
            event.append_to_arg_count(c)

    for c in '0123456789':
        handle_digit(c)

    @handle(Keys.Escape, '-')
    def _(event):
        """
        """
        if event._arg is None:
            event.append_to_arg_count('-')

    @handle(Keys.Escape, Keys.ControlJ, filter= ~has_selection)
    def _(event):
        """
        Meta + Newline: always accept input.
        """
        b = event.current_buffer

        if b.returnable(event.cli) and b.validate():
            b.add_to_history()
            event.cli.set_return_value(b.document)

        # Not returnable, but multiline.
        elif b.is_multiline:
            b.insert_text('\n')

    @handle(Keys.ControlSquareClose, Keys.Any)
    def _(event):
        """
        When Ctl-] + a character is pressed. go to that character.
        """
        match = event.current_buffer.document.find(event.data, in_current_line=True, count=(event.arg))
        if match is not None:
            event.current_buffer.cursor_position += match

    @handle(Keys.Escape, Keys.Backspace, filter= ~has_selection)
    def _(event):
        """
        Delete word backwards.
        """
        buffer = event.current_buffer
        pos = buffer.document.find_start_of_previous_word(count=event.arg)

        if pos:
            deleted = buffer.delete_before_cursor(count=-pos)
            event.cli.clipboard.set_text(deleted)

    @handle(Keys.Escape, 'a', filter= ~has_selection)
    def _(event):
        """
        Previous sentence.
        """
        # TODO:
        pass

    @handle(Keys.Escape, 'c', filter= ~has_selection)
    def _(event):
        """
        Capitalize the current (or following) word.
        """
        buffer = event.current_buffer

        for i in range(event.arg):
            pos = buffer.document.find_next_word_ending()
            words = buffer.document.text_after_cursor[:pos]
            buffer.insert_text(words.title(), overwrite=True)

    @handle(Keys.Escape, 'd', filter= ~has_selection)
    def _(event):
        """
        Delete word forwards.
        """
        buffer = event.current_buffer
        pos = buffer.document.find_next_word_ending(count=event.arg)

        if pos:
            deleted = buffer.delete(count=pos)
            event.cli.clipboard.set_text(deleted)

    @handle(Keys.Escape, 'e', filter= ~has_selection)
    def _(event):
        """ Move to end of sentence. """
        # TODO:
        pass

    @handle(Keys.Escape, 'f')
    @handle(Keys.ControlRight)
    def _(event):
        """
        Cursor to end of next word.
        """
        buffer= event.current_buffer
        pos = buffer.document.find_next_word_ending(count=event.arg)

        if pos:
            buffer.cursor_position += pos

    @handle(Keys.Escape, 'b')
    @handle(Keys.ControlLeft)
    def _(event):
        """
        Cursor to start of previous word.
        """
        buffer = event.current_buffer
        pos = buffer.document.find_previous_word_beginning(count=event.arg)
        if pos:
            buffer.cursor_position += pos

    @handle(Keys.Escape, 'l', filter= ~has_selection)
    def _(event):
        """
        Lowercase the current (or following) word.
        """
        buffer = event.current_buffer

        for i in range(event.arg):  # XXX: not DRY: see meta_c and meta_u!!
            pos = buffer.document.find_next_word_ending()
            words = buffer.document.text_after_cursor[:pos]
            buffer.insert_text(words.lower(), overwrite=True)

    @handle(Keys.Escape, 't', filter= ~has_selection)
    def _(event):
        """
        Swap the last two words before the cursor.
        """
        # TODO

    @handle(Keys.Escape, 'u', filter= ~has_selection)
    def _(event):
        """
        Uppercase the current (or following) word.
        """
        buffer = event.current_buffer

        for i in range(event.arg):
            pos = buffer.document.find_next_word_ending()
            words = buffer.document.text_after_cursor[:pos]
            buffer.insert_text(words.upper(), overwrite=True)

    @handle(Keys.Escape, '.', filter= ~has_selection)
    def _(event):
        """
        Rotate through the last word (white-space delimited) of the previous lines in history.
        """
        # TODO

    @handle(Keys.Escape, '\\', filter= ~has_selection)
    def _(event):
        """
        Delete all spaces and tabs around point.
        (delete-horizontal-space)
        """

    @handle(Keys.Escape, '*', filter= ~has_selection)
    def _(event):
        """
        `meta-*`: Insert all possible completions of the preceding text.
        """

    @handle(Keys.ControlX, Keys.ControlU, save_before=False, filter= ~has_selection)
    def _(event):
        event.current_buffer.undo()

    @handle(Keys.ControlX, 'o')
    def _(event):
        """
        Focus next buffer.
        """
        focus_next_buffer(event.cli)

    @handle(Keys.ControlX, Keys.ControlX)
    def _(event):
        """
        Move cursor back and forth between the start and end of the current
        line.
        """
        buffer = event.current_buffer

        if buffer.document.current_char == '\n':
            buffer.cursor_position += buffer.document.get_start_of_line_position(after_whitespace=False)
        else:
            buffer.cursor_position += buffer.document.get_end_of_line_position()

    @handle(Keys.ControlSpace)
    def _(event):
        """
        Start of the selection.
        """
        # Take the current cursor position as the start of this selection.
        event.current_buffer.start_selection(selection_type=SelectionType.CHARACTERS)

    @handle(Keys.ControlG, filter= ~has_selection)
    def _(event):
        """
        Control + G: Cancel completion menu and validation state.
        """
        event.current_buffer.complete_state = None
        event.current_buffer.validation_error = None

    @handle(Keys.ControlG, filter=has_selection)
    def _(event):
        """
        Cancel selection.
        """
        event.current_buffer.exit_selection()

    @handle(Keys.ControlW, filter=has_selection)
    @handle(Keys.ControlX, 'r', 'k', filter=has_selection)
    def _(event):
        """
        Cut selected text.
        """
        data = event.current_buffer.cut_selection()
        event.cli.clipboard.set_data(data)

    @handle(Keys.Escape, 'w', filter=has_selection)
    def _(event):
        """
        Copy selected text.
        """
        data = event.current_buffer.copy_selection()
        event.cli.clipboard.set_data(data)

    @handle(Keys.Escape, '<', filter= ~has_selection)
    def _(event):
        """
        Move to the first line in the history.
        """
        event.current_buffer.go_to_history(0)

    @handle(Keys.Escape, '>', filter= ~has_selection)
    def _(event):
        """
        Move to the end of the input history.
        This is the line we are editing.
        """
        buffer = event.current_buffer
        buffer.go_to_history(len(buffer._working_lines) - 1)

    @handle(Keys.Escape, Keys.Left)
    def _(event):
        """
        Cursor to start of previous word.
        """
        buffer = event.current_buffer
        buffer.cursor_position += buffer.document.find_previous_word_beginning(count=event.arg) or 0

    @handle(Keys.Escape, Keys.Right)
    def _(event):
        """
        Cursor to start of next word.
        """
        buffer = event.current_buffer
        buffer.cursor_position += buffer.document.find_next_word_beginning(count=event.arg) or \
            buffer.document.end_position

    @handle(Keys.Escape, '/', filter= ~has_selection)
    def _(event):
        """
        M-/: Complete.
        """
        event.current_buffer.complete_next()

    @handle(Keys.ControlC, '>', filter=has_selection)
    def _(event):
        """
        Indent selected text.
        """
        buffer = event.current_buffer

        buffer.cursor_position += buffer.document.get_start_of_line_position(after_whitespace=True)

        from_, to = buffer.document.selection_range()
        from_, _ = buffer.document.translate_index_to_position(from_)
        to, _ = buffer.document.translate_index_to_position(to)

        indent(buffer, from_ - 1, to, count=event.arg)  # XXX: why does translate_index_to_position return 1-based indexing???

    @handle(Keys.ControlC, '<', filter=has_selection)
    def _(event):
        """
        Unindent selected text.
        """
        buffer = event.current_buffer

        from_, to = buffer.document.selection_range()
        from_, _ = buffer.document.translate_index_to_position(from_)
        to, _ = buffer.document.translate_index_to_position(to)

        unindent(buffer, from_ - 1, to, count=event.arg)


def load_emacs_open_in_editor_bindings(registry, filter=None):
    """
    Pressing C-X C-E will open the buffer in an external editor.
    """
    handle = create_handle_decorator(registry, filter)
    has_selection = filters.HasSelection()

    @handle(Keys.ControlX, Keys.ControlE, filter= ~has_selection)
    def _(event):
        """
        Open editor.
        """
        event.current_buffer.open_in_editor()


def load_emacs_system_bindings(registry, filter=None, system_buffer_name='system'):
    handle = create_handle_decorator(registry, filter)
    has_focus = filters.HasFocus(system_buffer_name)

    @handle(Keys.Escape, '!', filter= ~has_focus)
    def _(event):
        """
        M-'!' opens the system prompt.
        """
        event.cli.focus_stack.push(system_buffer_name)

    @handle(Keys.Escape, filter=has_focus)
    @handle(Keys.ControlG, filter=has_focus)
    @handle(Keys.ControlC, filter=has_focus)
    def _(event):
        """
        Cancel system prompt.
        """
        event.cli.buffers[system_buffer_name].reset()
        event.cli.focus_stack.pop()

    @handle(Keys.ControlJ, filter=has_focus)
    def _(event):
        """
        Run system command.
        """
        system_line = event.cli.buffers[system_buffer_name]
        event.cli.run_system_command(system_line.text)
        system_line.reset(append_to_history=True)

        # Focus previous buffer again.
        event.cli.focus_stack.pop()


def load_emacs_search_bindings(registry, filter=None, search_buffer_name='search'):
    handle = create_handle_decorator(registry, filter)
    has_focus = filters.HasFocus(search_buffer_name)

    @handle(Keys.ControlG, filter=has_focus)
    # NOTE: the reason for not also binding Escape to this one, is that we want
    #       Alt+Enter to accept input directly in incremental search mode.
    def _(event):
        """
        Abort an incremental search and restore the original line.
        """
        search_line = event.cli.buffers[search_buffer_name]

        event.current_buffer.exit_isearch(restore_original_line=True)
        search_line.reset()
        event.cli.focus_stack.pop()

    @handle(Keys.ControlJ, filter=has_focus)
    def _(event):
        """
        When enter pressed in isearch, quit isearch mode. (Multiline
        isearch would be too complicated.)
        """
        search_line = event.cli.buffers[search_buffer_name]
        search_line.reset()

        event.cli.buffers[event.cli.focus_stack.previous].exit_isearch()
        event.cli.focus_stack.pop()

    @handle(Keys.ControlR, filter= ~has_focus)
    def _(event):
        event.cli.focus_stack.push(search_buffer_name)

        buffer = event.cli.buffers[event.cli.focus_stack.previous]
        buffer.incremental_search(IncrementalSearchDirection.BACKWARD)

    @handle(Keys.ControlS, filter= ~has_focus)
    def _(event):
        event.cli.focus_stack.push(search_buffer_name)

        buffer = event.cli.buffers[event.cli.focus_stack.previous]
        buffer.incremental_search(IncrementalSearchDirection.FORWARD)

    @handle(Keys.ControlR, filter=has_focus)
    @handle(Keys.Up, filter=has_focus)
    def _(event):
        buffer = event.cli.buffers[event.cli.focus_stack.previous]
        buffer.incremental_search(IncrementalSearchDirection.BACKWARD)

    @handle(Keys.ControlS, filter=has_focus)
    @handle(Keys.Down, filter=has_focus)
    def _(event):
        buffer = event.cli.buffers[event.cli.focus_stack.previous]
        buffer.incremental_search(IncrementalSearchDirection.FORWARD)



"""
Win32 asyncio event loop.

Windows notes:
- Somehow it doesn't seem to work with the 'ProactorEventLoop'.
"""
from __future__ import unicode_literals

from .asyncio_base import BaseAsyncioEventLoop
from ..terminal.vt100_input import InputStream

import os


__all__ = (
    'Win32AsyncioEventLoop',
)


class PosixAsyncioEventLoop(BaseAsyncioEventLoop):
    def __init__(self, input_processor, stdin, loop=None):
        super(PosixAsyncioEventLoop, self).__init__(input_processor, stdin)

        self._inputstream = InputStream(self.input_processor)

        # Create incremental decoder for decoding stdin.
        # We can not just do `os.read(stdin.fileno(), 1024).decode('utf-8')`, because
        # it could be that we are in the middle of a utf-8 byte sequence.
        self._stdin_decoder = self.stdin_decoder_cls()

    def wait_for_input(self, f_ready):
        def stdin_ready():
            data = self._read_from_stdin()
            self._inputstream.feed_and_flush(data)

            f_ready.set_result(None)  # Quit coroutine
            self.loop.remove_reader(self.stdin.fileno())

        self.loop.add_reader(self.stdin.fileno(), stdin_ready)

    def _read_from_stdin(self):
        """
        Read the input and return it.
        """
        # Note: the following works better than wrapping `self.stdin` like
        #       `codecs.getreader('utf-8')(stdin)` and doing `read(1)`.
        #       Somehow that causes some latency when the escape
        #       character is pressed. (Especially on combination with the `select`.
        try:
            bytes = os.read(self.stdin.fileno(), 1024)
        except OSError:
            # In case of SIGWINCH
            bytes = b''

        try:
            return self._stdin_decoder.decode(bytes)
        except UnicodeDecodeError:
            # When it's not possible to decode this bytes, reset the decoder.
            # The only occurence of this that I had was when using iTerm2 on OS
            # X, with "Option as Meta" checked (You should choose "Option as
            # +Esc".)
            self._stdin_decoder = self.stdin_decoder_cls()
            return ''



from __future__ import unicode_literals

from pygments.lexers import BashLexer
from pygments.token import Token

from ..enums import IncrementalSearchDirection

from .processors import BeforeInput

from . import Window
from .dimension import LayoutDimension
from .controls import BufferControl, TokenListControl, UIControl
from .utils import token_list_len
from .screen import Screen
from prompt_toolkit.filters import HasFocus, HasArg, HasCompletions, HasValidationError, HasSearch, AlwaysOff, AlwaysOn, IsDone


__all__ = (
    'ArgToolbar',
    'CompletionsToolbar',
    'SearchToolbar',
    'SystemToolbar',
    'ValidationToolbar',
)


class TokenListToolbar(Window):
    def __init__(self, get_tokens, default_char=None, filter=None):
        super(TokenListToolbar, self).__init__(
            TokenListControl(get_tokens, default_char=default_char),
            height=LayoutDimension.exact(1),
            filter=filter)


class SystemToolbarControl(BufferControl):
    def __init__(self):
        super(SystemToolbarControl, self).__init__(
            lexer=BashLexer,
            buffer_name='system',
            show_line_numbers=AlwaysOff(),
            input_processors=[BeforeInput.static('Shell command: ', Token.Toolbar.System.Prefix)],)


class SystemToolbar(Window):
    def __init__(self):
        super(SystemToolbar, self).__init__(
            SystemToolbarControl(),
            height=LayoutDimension.exact(1),
            filter=HasFocus('system') & ~IsDone())


class ArgToolbarControl(TokenListControl):
    def __init__(self):
        def get_tokens(cli):
            return [
                (Token.Toolbar.Arg, 'Repeat: '),
                (Token.Toolbar.Arg.Text, str(cli.input_processor.arg)),
            ]

        super(ArgToolbarControl, self).__init__(get_tokens)


class ArgToolbar(Window):
    def __init__(self):
        super(ArgToolbar, self).__init__(
            ArgToolbarControl(),
            height=LayoutDimension.exact(1),
            filter=HasArg())


class SearchToolbarControl(BufferControl):
    def __init__(self):
        token = Token.Toolbar.Search

        def get_before_input(cli, buffer):
            buffer = cli.buffers[cli.focus_stack.previous]

            if buffer.isearch_state is None:
                text = ''
            elif buffer.isearch_state.isearch_direction == IncrementalSearchDirection.BACKWARD:
                text = 'I-search backward: '
            else:
                text = 'I-search: '

            return [(token, text)]

        super(SearchToolbarControl, self).__init__(
            buffer_name='search',
            input_processors=[BeforeInput(get_before_input)],
        )


class SearchToolbar(Window):
    def __init__(self):
        super(SearchToolbar, self).__init__(
            SearchToolbarControl(),
            height=LayoutDimension.exact(1),
            filter=HasSearch() & ~IsDone())


class CompletionsToolbarControl(UIControl):
    token = Token.Toolbar.Completions

    def create_screen(self, cli, width, height):
        complete_state = cli.current_buffer.complete_state
        if complete_state:
            completions = complete_state.current_completions
            index = complete_state.complete_index  # Can be None!

            # Width of the completions without the left/right arrows in the margins.
            content_width = width - 6

            # Booleans indicating whether we stripped from the left/right
            cut_left = False
            cut_right = False

            # Create Menu content.
            tokens = []

            for i, c in enumerate(completions):
                # When there is no more place for the next completion
                if token_list_len(tokens) + len(c.display) >= content_width:
                    # If the current one was not yet displayed, page to the next sequence.
                    if i <= (index or 0):
                        tokens = []
                        cut_left = True
                    # If the current one is visible, stop here.
                    else:
                        cut_right = True
                        break

                tokens.append((self.token.Completion.Current if i == index else self.token.Completion, c.display))
                tokens.append((self.token, ' '))

            # Extend/strip until the content width.
            tokens.append((self.token, ' ' * (content_width - len(tokens))))
            tokens = tokens[:content_width]

            # Return tokens
            all_tokens = [
                (self.token, ' '),
                (self.token.Arrow, '<' if cut_left else ' '),
                (self.token, ' '),
            ] + tokens + [
                (self.token, ' '),
                (self.token.Arrow, '>' if cut_right else ' '),
                (self.token, ' '),
            ]
        else:
            all_tokens = []

        screen = Screen(width)
        screen.write_at_position(all_tokens, width)
        return screen


class CompletionsToolbar(Window):
    def __init__(self, extra_filter=AlwaysOn()):
        super(CompletionsToolbar, self).__init__(
            CompletionsToolbarControl(),
            height=LayoutDimension.exact(1),
            filter=HasCompletions() & ~IsDone() & extra_filter)


class ValidationToolbarControl(TokenListControl):
    def __init__(self):
        token = Token.Toolbar.Validation

        def get_tokens(cli):
            buffer = cli.current_buffer

            if buffer.validation_error:
                row, column = buffer.document.translate_index_to_position(
                    buffer.validation_error.index)

                text = '%s (line=%s column=%s)' % (
                    buffer.validation_error.message, row, column)
                return [(token, text)]
            else:
                return []

        super(ValidationToolbarControl, self).__init__(get_tokens)


class ValidationToolbar(Window):
    def __init__(self):
        super(ValidationToolbar, self).__init__(
            ValidationToolbarControl(),
            height=LayoutDimension.exact(1),
            filter=HasValidationError() & ~IsDone())



from __future__ import unicode_literals

import six
from collections import defaultdict

try:
    from wcwidth import wcwidth
except ImportError:
    from .libs.wcwidth import wcwidth


__all__ = (
    'EventHook',
    'DummyContext',
)


class EventHook(object):
    """
    Event hook::

        e = EventHook()
        e += handler_function  # Add event handler.
        e.fire()  # Fire event.

    Thanks to Michael Foord:
    http://www.voidspace.org.uk/python/weblog/arch_d7_2007_02_03.shtml#e616
    """
    def __init__(self):
        self.__handlers = []

    def __iadd__(self, handler):
        self.__handlers.append(handler)
        return self

    def __isub__(self, handler):
        self.__handlers.remove(handler)
        return self

    def fire(self, *args, **keywargs):
        for handler in self.__handlers:
            handler(*args, **keywargs)


class DummyContext(object):
    """
    (contextlib.nested is not available on Py3)
    """
    def __enter__(self):
        pass

    def __exit__(self, *a):
        pass


class _CharSizesCache(dict):
    """
    Cache for wcwidth sizes.
    """
    def __missing__(self, string):
        # Note: We use the `max(0, ...` because some non printable control
        #       characters, like e.g. Ctrl-underscore get a -1 wcwidth value.
        #       It can be possible that these characters end up in the input
        #       text.
        if len(string) == 1:
            result = max(0, wcwidth(string))
        else:
            result = sum(max(0, wcwidth(c)) for c in string)

        self[string] = result
        return result


_CHAR_SIZES_CACHE = _CharSizesCache()


def get_cwidth(string):
    """
    Return width of a string. Wrapper around ``wcwidth``.
    """
    return _CHAR_SIZES_CACHE[string]



"""
The main `CommandLineInterface` class and logic.
"""
from __future__ import unicode_literals

import os
import signal
import six
import sys
import weakref

from .buffer import Buffer
from .clipboard import Clipboard
from .completion import CompleteEvent
from .focus_stack import FocusStack
from .history import History
from .key_binding.bindings.emacs import load_emacs_bindings
from .key_binding.input_processor import InputProcessor
from .key_binding.registry import Registry
from .layout import Window
from .layout.controls import BufferControl
from .renderer import Renderer
from .utils import EventHook, DummyContext
from .filters import AlwaysOff, AlwaysOn

from pygments.styles.default import DefaultStyle
from types import GeneratorType

if sys.platform == 'win32':
    from .terminal.win32_input import raw_mode, cooked_mode
    from .eventloop.win32 import Win32EventLoop as EventLoop
else:
    from .terminal.vt100_input import raw_mode, cooked_mode
    from .eventloop.posix import PosixEventLoop as EventLoop
    from .eventloop.posix import call_on_sigwinch


__all__ = (
    'AbortAction',
    'Exit',
    'Abort',
    'CommandLineInterface',
)


class AbortAction:
    """
    Actions to take on an Exit or Abort exception.
    """
    IGNORE = 'ignore'
    RETRY = 'retry'
    RAISE_EXCEPTION = 'raise-exception'
    RETURN_NONE = 'return-none'


class Exit(Exception):
    pass


class Abort(Exception):
    pass


class CommandLineInterface(object):
    """
    Wrapper around all the other classes, tying everything together.

    Typical usage::

        cli = CommandLineInterface()
        while True:
            result = cli.read_input()
            print(result)

    :param stdin: Input stream, by default sys.stdin
    :param stdout: Output stream, by default sys.stdout
    :param layout: :class:`Layout` instance.
    :param style: :class:`Layout` instance.
    :param create_async_autocompleters: Boolean. If True, autocompletions will
        be generated asynchronously while you type.
    """
    def __init__(self, stdin=None, stdout=None,
                 layout=None,
                 buffer=None,
                 buffers=None,
                 style=None,
                 key_bindings_registry=None,
                 clipboard=None,
                 create_async_autocompleters=True,
                 renderer=None,
                 initial_focussed_buffer='default'):

        assert buffer is None or isinstance(buffer, Buffer)
        assert buffers is None or isinstance(buffers, dict)

        self.stdin = stdin or sys.__stdin__
        self.stdout = stdout or sys.__stdout__
        self.style = style or DefaultStyle
        self.create_async_autocompleters = create_async_autocompleters

        # Events

        #: Called when there is no input for x seconds.
        #: (Fired when any eventloop.onInputTimeout is fired.)
        self.onInputTimeout = EventHook()

        self.onReadInputStart = EventHook()
        self.onReadInputEnd = EventHook()
        self.onReset = EventHook()

        # Focus stack.
        self.focus_stack = FocusStack(initial=initial_focussed_buffer)

        #: The input buffers.
        self.buffers = {
            # We don't make the 'search' and 'system' focussable, that means
            # that window-focus changing functionality is disabled for these.
            # (See prompt_toolkit.key_binding.bindings.utils.focus_next_buffer.)
            # Also, 'returable' is False, in order to block normal Enter/ControlC behaviour.
            'default': (buffer or Buffer(returnable=AlwaysOn())),
            'search': Buffer(history=History(), focussable=AlwaysOff(), returnable=AlwaysOff()),
            'system': Buffer(history=History(), focussable=AlwaysOff(), returnable=AlwaysOff()),
        }
        if buffers:
            self.buffers.update(buffers)

        # When the text in the search buffer changes, set search string in
        # other buffers.
        def search_text_changed():
            search_text = self.buffers['search'].text

            for name, b in self.buffers.items():
                if name != 'search':
                    b.set_search_text(search_text)

        self.buffers['search'].onTextChanged += search_text_changed

        #: The `Layout` instance.
        self.layout = layout or Window(BufferControl())

        #: The clipboard instance
        self.clipboard = clipboard or Clipboard()

        #: The `Renderer` instance.
        # Make sure that the same stdout is used, when a custom renderer has been passed.
        self.renderer = renderer or Renderer(stdout=self.stdout)
        assert self.renderer.stdout == self.stdout

        if key_bindings_registry is None:
            key_bindings_registry = Registry()
            load_emacs_bindings(key_bindings_registry)

        #: The `InputProcessor` instance.
        self.input_processor = InputProcessor(key_bindings_registry, weakref.ref(self))

        # Handle events.
        if self.create_async_autocompleters:
            for b in self.buffers.values():
                if b.completer:
                    b.onTextInsert += self._create_async_completer(b)

        self._reset()

        # Event loop.
        self.eventloop = None

    @property
    def is_reading_input(self):
        return bool(self.eventloop)

    @property
    def current_buffer_name(self):
        """
        The name of the current  :class:`Buffer`.
        """
        return self.focus_stack.current

    @property
    def current_buffer(self):
        """
        The current focussed :class:`Buffer`.
        """
        return self.buffers[self.focus_stack.current]

    def add_buffer(self, name, buffer, focus=False):
        """
        Insert a new buffer.
        """
        self.buffers[name] = buffer

        if self.create_async_autocompleters:
            if buffer.completer:
                buffer.onTextInsert += self._create_async_completer(buffer)

        if focus:
            self.focus_stack.replace(name)

    def _reset(self):
        """
        Reset everything, for reading the next input.
        """
        # Notice that we don't reset the buffers. (This happens just before
        # returning, and when we have multiple buffers, we clearly want the
        # content in the other buffers to remain unchanged between several
        # calls of `read_input`. (And the same is true for the `focus_stack`.)

        self._exit_flag = False
        self._abort_flag = False
        self._return_value = None

        self.renderer.reset()
        self.input_processor.reset()
        self.layout.reset()

        # Trigger reset event.
        self.onReset.fire()

    def request_redraw(self):
        """
        Thread safe way of sending a repaint trigger to the input event loop.
        """
        if self.is_reading_input:
            self.call_from_executor(self._redraw)

    def _redraw(self):
        """
        Render the command line again. (Not thread safe!)
        (From other threads, or if unsure, use `request_redraw`.)
        """
        self.renderer.render(self, self.layout, self.style)

    def run_in_executor(self, callback):
        """
        Run a long running function in a background thread.
        (This is recommended for code that could block the `read_input` event
        loop.)
        Similar to Twisted's ``deferToThread``.

        :param callback: The callable that should run in the executor.
        """
        self.eventloop.run_in_executor(callback)

    def call_from_executor(self, callback):
        """
        Call this function in the main event loop.
        Similar to Twisted's ``callFromThread``.

        :param callback: The callable that should run in the main event loop.
        """
        if self.eventloop:
            self.eventloop.call_from_executor(callback)
            return True
        else:
            return False

    def _on_resize(self):
        """
        When the window size changes, we erase the current output and request
        again the cursor position. We the CPR answer arrives, the output is
        drawn again.
        (We do it asynchronously, because writing to the output from inside the
        signal handler causes easily reentrant calls, giving runtime errors.)
        """
        assert self.eventloop

        def do_in_event_loop():
            # Erase, request position (when cursor is at the start position)
            # and redraw again. -- The order is important.
            self.renderer.erase()
            self.renderer.request_absolute_cursor_position()
            self._redraw()
        self.call_from_executor(do_in_event_loop)

    def read_input(self, on_abort=AbortAction.RETRY, on_exit=AbortAction.IGNORE):
        """
        Read input string from command line.

        :param on_abort: :class:`AbortAction` value. What to do when Ctrl-C has been pressed.
        :param on_exit:  :class:`AbortAction` value. What to do when Ctrl-D has been pressed.
        """
        eventloop = EventLoop(self.input_processor, self.stdin)
        g = self._read_input(on_abort=on_abort,
                             on_exit=on_exit,
                             eventloop=eventloop,
                             return_corountine=False)

        # Return result from `_read_input`.
        try:
            while True:
                next(g)
        except StopIteration as e:
            # Clean up renderer. (This will leave the alternate screen, if we
            # use that.)
            self.renderer.reset()

            # Return result.
            return e.args[0]

    def read_input_async(self,
                         on_abort=AbortAction.RETRY, on_exit=AbortAction.IGNORE):
        """
        Same as `read_input`, but this returns an asyncio coroutine.

        Warning: this will only work on Python >3.3
        """
        # Inline import, to make sure the rest doesn't break on Python 2
        if sys.platform == 'win32':
            from prompt_toolkit.eventloop.asyncio_win32 import Win32AsyncioEventLoop as AsyncioEventLoop
        else:
            from prompt_toolkit.eventloop.asyncio_posix import PosixAsyncioEventLoop as AsyncioEventLoop

        eventloop = AsyncioEventLoop(self.input_processor, self.stdin)
        return self._read_input(on_abort=on_abort,
                                on_exit=on_exit,
                                eventloop=eventloop,
                                return_corountine=True)

    def _read_input(self, on_abort, on_exit,
                    eventloop, return_corountine):
        """
        The implementation of ``read_input`` which can be called both
        synchronously and asynchronously. When called as coroutine it will
        delegate all the futures from ``eventloop.loop_coroutine``. In both
        cases it returns the result through ``StopIteration``.
        """
        # Set `is_reading_input` flag.
        if self.is_reading_input:
            raise Exception('Already reading input. Read_input is not thread safe.')

        # Create new event loop.
        self.eventloop = eventloop
        self.eventloop.onInputTimeout += lambda: self.onInputTimeout.fire()

        try:
            def reset():
                self._reset()
            reset()

            # Trigger onReadInputStart event.
            self.onReadInputStart.fire()

            with raw_mode(self.stdin.fileno()):
                self.renderer.request_absolute_cursor_position()
                self._redraw()

                with (DummyContext() if sys.platform == 'win32' else
                      call_on_sigwinch(self._on_resize)):
                    while True:
                        if return_corountine:
                            loop_result = self.eventloop.loop_coroutine()
                            assert isinstance(loop_result, GeneratorType)

                            for future in loop_result:
                                yield future
                        else:
                            loop_result = self.eventloop.loop()
                            assert not isinstance(loop_result, GeneratorType)

                        # If the exit flag has been set.
                        if self._exit_flag:
                            self.current_buffer.reset()
                            if on_exit != AbortAction.IGNORE:
                                self._redraw()
                                self.current_buffer.reset()

                            if on_exit == AbortAction.RAISE_EXCEPTION:
                                raise Exit()
                            elif on_exit == AbortAction.RETURN_NONE:
                                raise StopIteration(None)
                            elif on_exit == AbortAction.RETRY:
                                reset()
                                self.renderer.request_absolute_cursor_position()

                        # If the abort flag has been set.
                        if self._abort_flag:
                            if on_abort != AbortAction.IGNORE:
                                self._redraw()
                                self.current_buffer.reset()

                            if on_abort == AbortAction.RAISE_EXCEPTION:
                                raise Abort()
                            elif on_abort == AbortAction.RETURN_NONE:
                                raise StopIteration(None)
                            elif on_abort == AbortAction.RETRY:
                                reset()
                                self.renderer.request_absolute_cursor_position()

                        # If a return value has been set.
                        if self._return_value is not None:
                            self._redraw()
                            self.current_buffer.reset()
                            raise StopIteration(self._return_value)

                        # Now render the current layout to the output.
                        self._redraw()
        finally:
            # Close event loop
            self.eventloop.close()
            self.eventloop = None

            # Trigger onReadInputEnd event.
            self.onReadInputEnd.fire()

    def set_exit(self):
        self._exit_flag = True

    def set_abort(self):
        self._abort_flag = True

    def set_return_value(self, document):
        self._return_value = document

    def run_in_terminal(self, func):
        """
        Run function on the terminal above the prompt.

        What this does is first hiding the prompt, then running this callable
        (which can safely output to the terminal), and then again rendering the
        prompt which causes the output of this function to scroll above the
        prompt.

        This function is thread safe. It will return immediately and the
        callable will be scheduled in the eventloop of this
        :class:`CommandLineInterface`.
        """
        self.call_from_executor(lambda: self._run_in_terminal(func))

    def _run_in_terminal(self, func):
        """ Blocking, not thread safe version of ``run_in_terminal``. """
        assert self.is_reading_input, 'Should be called while reading input.'
        self.renderer.erase()

        # Run system command.
        with cooked_mode(self.stdin.fileno()):
            func()

        self.renderer.reset()
        self.renderer.request_absolute_cursor_position()
        self._redraw()

    def run_system_command(self, command):
        """
        Run system command (While hiding the prompt. When finished, all the
        output will scroll above the prompt.)

        :param command: Shell command to be executed.
        """
        def run():
            if sys.platform == 'win32':
                os.system(command)  # Needs to be unicode for win32
            else:
                os.system(command.encode('utf-8'))

            (input if six.PY3 else raw_input)('\nPress ENTER to continue...')

        self.run_in_terminal(run)

    def suspend_to_background(self):
        """
        (Not thread safe -- to be called from inside the key bindings.)
        Suspend process.
        """
        def run():
            # Send `SIGSTP` to own process.
            # This will cause it to suspend.
            os.kill(os.getpid(), signal.SIGTSTP)

        self.run_in_terminal(run)

    @property
    def is_exiting(self):
        """
        ``True`` when the exit flag as been set.
        """
        return self._exit_flag

    @property
    def is_aborting(self):
        """
        ``True`` when the abort flag as been set.
        """
        return self._abort_flag

    @property
    def is_returning(self):
        """
        ``True`` when a return value has been set.
        """
        return self._return_value is not None

    @property
    def is_done(self):
        return self.is_exiting or self.is_aborting or self.is_returning

    def _create_async_completer(self, buffer):
        """
        Create function for asynchronous autocompletion while typing.
        (Autocomplete in other thread.)
        """
        complete_thread_running = [False]  # By ref.

        def async_completer():
            document = buffer.document

            # Don't start two threads at the same time.
            if complete_thread_running[0]:
                return

            # Don't complete when we already have completions.
            if buffer.complete_state:
                return

            # Don't automatically complete on empty inputs.
            if not buffer.text:
                return

            # Otherwise, get completions in other thread.
            complete_thread_running[0] = True

            def run():
                completions = list(buffer.completer.get_completions(
                    document,
                    CompleteEvent(text_inserted=True)))
                complete_thread_running[0] = False

                def callback():
                    """
                    Set the new complete_state in a safe way. Don't replace an
                    existing complete_state if we had one. (The user could have
                    pressed 'Tab' in the meantime. Also don't set it if the text
                    was changed in the meantime.
                    """
                    # Set completions if the text was not yet changed.
                    if buffer.text == document.text and \
                            buffer.cursor_position == document.cursor_position and \
                            not buffer.complete_state:
                        buffer._start_complete(go_to_first=False, completions=completions)
                        self._redraw()
                    else:
                        # Otherwise, restart thread.
                        async_completer()
                self.call_from_executor(callback)

            self.run_in_executor(run)
        return async_completer

    def stdout_proxy(self):
        """
        Create an :class:`_StdoutProxy` class which can be used as a patch for
        sys.stdout. Writing to this proxy will make sure that the text appears
        above the prompt, and that it doesn't destroy the output from the
        renderer.
        """
        return _StdoutProxy(self)

    def patch_stdout_context(self):
        """
        Return a context manager that will replace ``sys.stdout`` with a proxy
        that makes sure that all printed text will appear above the prompt, and
        that it doesn't destroy the output from the renderer.
        """
        return _PatchStdoutContext(self.stdout_proxy())


class _PatchStdoutContext(object):
    def __init__(self, new_stdout):
        self.new_stdout = new_stdout

    def __enter__(self):
        self.original_stdout = sys.stdout
        sys.stdout = self.new_stdout

    def __exit__(self, *a, **kw):
        sys.stdout = self.original_stdout


class _StdoutProxy(object):
    """
    Proxy for stdout, as returned by
    :class:`CommandLineInterface.stdout_proxy`.
    """
    def __init__(self, cli):
        self._cli = cli
        self._buffer = []

    def _do(self, func):
        if self._cli.is_reading_input:
            self._cli.run_in_terminal(func)
        else:
            func()

    def write(self, data):
        """
        Note: print()-statements cause to multiple write calls.
              (write('line') and write('\n')). Of course we don't want to call
              `run_in_terminal` for every individual call, because that's too
              expensive, and as long as the newline hasn't been written, the
              text itself is again overwritter by the rendering of the input
              command line. Therefor, we have a little buffer which holds the
              text until a newline is written to stdout.
        """
        if '\n' in data:
            # When there is a newline in the data, write everything before the
            # newline, including the newline itself.
            before, after = data.rsplit('\n', 1)
            to_write = self._buffer + [before, '\n']
            self._buffer = [after]

            def run():
                for s in to_write:
                    self._cli.stdout.write(s)
            self._do(run)
        else:
            # Otherwise, cache in buffer.
            self._buffer.append(data)

    def flush(self):
        """
        Flush buffered output.
        """
        def run():
            for s in self._buffer:
                self._cli.stdout.write(s)
            self._buffer = []
            self._cli.stdout.flush()
        self._do(run)

    def __getattr__(self, name):
        return getattr(self._cli.stdout, name)
